/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
// pour getoptlong
#include <unistd.h>
#include <getopt.h>

#include <signal.h>
#include <setjmp.h>

#include "fmr.h"
#include "fmr_io.h"
#include "fmr_mem.h"
#include "fmr_tools.h"
#include "fmr_cost.h"
#include "fmr_core.h"
#include "fmr_constraints.h"
#include "fmr_servicegraph.h"
#include "fmr_composanteConnexe.h"
#include "fmr_glouton.h"
#include "fmr_stats.h"
#include "fmr_threads.h"
#include <pthread.h>

#define EQUIPE "S26"

/* La partie constante de l'instance est une variable globale */
instance_t instance;

/* pour setjmp */
jmp_buf timeout_jump;

volatile sig_atomic_t alarme;
void
catch_alarm (int sig)
{
    alarme = 1;
    longjmp (timeout_jump, 1);
}


typedef struct {
    placement_t * initial ;
    placement_t * initial_relax ;
    char * algos ;
    unsigned int seed;
} thread_data_t ;

void * threaded_calcul_assignement( void * data )
{
    pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
    calcul_assignement(((thread_data_t*)data)->initial, ((thread_data_t*)data)->initial_relax, ((thread_data_t*)data)->algos, ((thread_data_t*)data)->seed);
    return NULL;
}


char *output_file = NULL;
char *listeAlgosA = NULL;
char *listeAlgosB = NULL;
char *output_struct = NULL;
bool output_structure = false;
bool affiche_cout_final = false;

unsigned int time_limit = 300;     // defaut: 5mn

int
main (int argc, char **argv)
{
    char *instance_file = NULL;
    char *input_file = NULL;
    int rc, c;
    unsigned int maxcomp, maxcomp2;
    placement_t initial;
    placement_t initial_relax;
    stats_t *statistiques_placement;
    pthread_t threadA, threadB;
    thread_data_t thread_dataA = {NULL, NULL, NULL, 1}, thread_dataB = {NULL, NULL, NULL, 1} ;
    int affiche_equipe = false ;
    int nb_options = 0 ;
    
    //    graphnode_t* graphTab; /* tableau qui contient le graphe de dependances */

    /* Parsing des arguments */
    static struct option long_options[] = {
	{"h", no_argument, NULL, 'h'},	//help
	{"?", no_argument, NULL, 'h'},	//help
	{"t", required_argument, NULL, 't'},	//time_limit en secondes
	{"p", required_argument, NULL, 'p'},	//instance filename
	{"i", required_argument, NULL, 'i'},	//initial solution filename
	{"o", required_argument, NULL, 'o'},	// new solution filename
	{"s", required_argument, NULL, 's'},	//seed
	{"name", no_argument, NULL, 'n'},	/* nom de l'équipe * les options qui suivent sont perso ! */
	{"v", required_argument, NULL, 'v'},	//output structure resultat
	{"A", required_argument, NULL, 'A'},	//Algorithmes a utiliser dans le premier thread
	{"B", required_argument, NULL, 'B'},	//Algorithmes a utiliser dans le second thread
	{"c", no_argument, NULL, 'c'},	// affiche le coût final avant de sortir
	{0, 0, 0, 0}
    };


    /* getopt_long stores the option index here. */
    int option_index = 0;
     thread_dataA.algos = "GSCV";
     thread_dataB.algos = "BCN";
    output_file = "out";	/* valeur par défaut du fichier de sortie */


    /* Initialisations pour la minuterie */
    tempsInit2();
    signal (SIGALRM, catch_alarm);

    while (1) {
	c = getopt_long_only (argc, argv, "", long_options, &option_index);
	/* Detect the end of the options. */
	if (c == -1)
	    break;
#define MESSAGE "options supportées : -h -t (time_limit) -p (instance_filename) " \
      "-i (initial solution filename) -o (new solution filename) -s (seed) -name\n"
      
        nb_options ++ ;
	switch (c) {
	case 'h':
	    printf (MESSAGE);
	    return EXIT_SUCCESS;
	case 't':
	    time_limit = (unsigned int) strtoul (optarg, NULL, 10);
	    DEBUG_IO_PRINT ("time limit=%d s\n", time_limit);
	    break;
	case 'p':
	    instance_file = optarg;
	    DEBUG_IO_PRINT ("Reading instance from file %s\n", instance_file);
	    break;
	case 'i':
	    input_file = optarg;
	    DEBUG_IO_PRINT ("Reading init from file %s\n", input_file);
	    break;
	case 'o':
	    output_file = optarg;
	    DEBUG_IO_PRINT ("The output will be written into file %s\n",
			    output_file);
	    break;
	case 's':
	    thread_dataA.seed = thread_dataB.seed = (unsigned int) strtol (optarg, NULL, 10);
	    break;
	case 'v':
	    output_struct = optarg;
	    output_structure = true;
	    break;
	case 'c':
	    affiche_cout_final = true;
	    break;
	case 'n':
	    printf("%s\n", EQUIPE);
	    affiche_equipe = true ;
	    break;
	case 'A':
	    thread_dataA.algos = optarg;
	    break;
	case 'B':
	    thread_dataB.algos = optarg;
	    break;

	default:
	    printf (MESSAGE), abort ();
	}
    }
    
    /* On sort sans erreur -name est la seule option */
    if (affiche_equipe && nb_options==1)
	return EXIT_SUCCESS;
    
    /* On fixe le temps limite, mainetenant qu'il est connu */
    if (time_limit>1)
	alarm (time_limit-1);

    /* On ouvre et on lit le fichier d'instance */
    rc = lire_instance (instance_file);
    if (0 > rc) {
	printf ("Erreur lecture fichier d\'instance\n");
	return EXIT_FAILURE;
    }

#ifdef DEBUGSTATS
    printf("Après lecture modele time=%lf\n",tempsEcoule2());
    printMemoireUtilisee();
#endif
    /* Allocation des structures de donnees contenant l'ordonnancement */

    initial = alloc_assignement (instance.nb_procs);
    //initial_relax = alloc_assignement (instance.nb_procs);

    /* On ouvre et on lit le fichier contenant la solution initiale */

    /*modif_sc();*/
    /*rc = lire_init (input_file, instance.nb_procs, &initial_relax);
    if (0 > rc) {
	fprintf (stderr, "Erreur lecture fichier d\'input\n");
	free_placement (&initial_relax);
	return EXIT_FAILURE;
    }*/
    /*unmodif_sc();*/
    rc = lire_init (input_file, instance.nb_procs, &initial);
    if (0 > rc) {
	fprintf (stderr, "Erreur lecture fichier d\'input\n");
	free_placement (&initial);
	return EXIT_FAILURE;
    }

#ifdef DEBUGSTATS
    printf("Après lecture solution init. time=%lf\n",tempsEcoule2());
    printMemoireUtilisee();

	check_depassements_capa (&initial);
	compConnexe (&maxcomp, &maxcomp2);
	printf("maxcomp=%u, maxcomp2=%u\n",maxcomp,maxcomp2);
    	printMemoireUtilisee();
#endif

    /* Construction du graphe de dependances des services
     *       Ce serait peut-etre bien de rendre cet appel optionnel
     */

    //graphTab = (graphnode_t*) malloc( instance.nb_services * sizeof( graphnode_t ) );
    //construireGraphDependances( graphTab );

    /* Calcul des stats du placement initial */

    //statistiques_placement = calcul_stats_placement (initial);
    //rc = afficher_stats (statistiques_placement);

    /* en cas de timeout on revient ici avec la valeur 1 sinon setjmp retourne 0 */
    if (setjmp (timeout_jump) == 0) {

	initial.cout = cout (&initial, &initial);

	DEBUG_PRINT ("init coute=%lld\n", initial.cout);
	atomic_init_meilleur_placement();
	atomic_set_meilleur_placement(&initial, '.');
	thread_dataA.initial = &initial ;
	thread_dataA.initial_relax = &initial_relax ;
	thread_dataB.initial = &initial ;
	thread_dataB.initial_relax = &initial_relax ;
	
	pthread_create(&threadA, NULL, threaded_calcul_assignement, &thread_dataA) ;
	pthread_create(&threadB, NULL, threaded_calcul_assignement, &thread_dataB) ;
	pthread_join(threadA, NULL) ;
	pthread_join(threadB, NULL) ;

    } else {
	DEBUG_STATS_PRINT(" timeout\n");
	DEBUG_PRINT (" timeout\n");
	pthread_cancel(threadA);
	pthread_cancel(threadB);
// 	DEBUG_PRINT (" threads cancelled (timeout)\n");
	pthread_join(threadA, NULL) ;
	pthread_join(threadB, NULL) ;
	// on verrouille le résultat sans s'occuper des threads qui continuent tranquillement
// 	atomic_lock_placement();
    }
    
    
    #ifdef DEBUGSTATS
    check_depassements_capa (atomic_get_meilleur_placement());
    DEBUG_PRINT("on est passe de %lld a %lld\n", initial.cout, atomic_get_meilleur_placement()->cout);
    if (output_structure)
	ecrire_struct_resultat (output_struct, atomic_get_meilleur_placement(), maxcomp,
				maxcomp2, initial.cout, atomic_get_meilleur_placement()->cout);
    printf("Avant ecriture resultat time=%lf\n",tempsEcoule2());
    printMemoireUtilisee();
    printf("memoire restante = %f Mo\n",(float)memoireRestante()/1E6);
    #endif
    
	/* Finalisation */

	ecrire_resultat (output_file, atomic_get_meilleur_placement());

#ifdef TRACE_ALGOS
	printf("algorithmes utilisés : %s\n\n", atomic_get_meilleur_placement()->liste_algos);
#endif
	if (affiche_cout_final)
	    fprintf (stdout, "%lld\n", atomic_get_meilleur_placement()->cout);

	atomic_free_placement();

	if (0 > rc) {
	    printf ("Erreur ecriture fichier resultat\n");
	    return EXIT_FAILURE;
	}
	return EXIT_SUCCESS;
}
