/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/

#ifndef _FMR_H_
#define _FMR_H_

/* includes utiles à tous */
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <stdbool.h>

/* Maxima définis dans le projet */
/* pour ceux<=5000 : tableaux préalloués pour gagner du temps (ou tous ?) */
#define MAXNBMACHINES	5000
#define MAXNBRESSOURCES   20
#define MAXNBPROCESSUS 50000
/* rectification des specs !!!! */
#define MAXNBSERVICES  50000
#define MAXNBNEIGHBORHOOD 1000
#define MAXNBLOCATIONS  1000
#define MAXNBBALANCECOSTS 10

typedef long long int int64;

struct ressource { 
    bool transcient;       	/* La ressource est-elle ephemere */
    unsigned weight_loadCost;   /* Poids (weight_loadCost) de la ressource */
};
typedef struct ressource ressource_t;

struct processus {
    unsigned short service;         /* Service auquel le processus appartient */
    unsigned* R;          	    /* R(pi,rk) requirementes: Besoins du processus dans les differentes ressources */
    unsigned PMC;                   /* Cout de deplacement du processus */
    /*** champs internes ***/
    unsigned short id;               /* Numero du processus: inutile ? */
};
typedef struct processus processus_t;

struct process_list {
    processus_t* process;
    struct process_list* next;
};

struct machine { 
    /*** champs obligatoires ***/
    unsigned* C;           /* Capacites des differentes ressources de la machine */
    unsigned* SC;          /* Safety Capacities : Capacites secu de la machine dans les ressources */ 
    //unsigned* SCsauv;          /* Safety Capacities sauvegarde: Capacites secu de la machine dans les ressources */ 
    unsigned short neighborhood;      /* numero de voisinage auquel la machine appartient */
    unsigned short location;          /* localisation de la machine */
    unsigned* MMC;         /* Machine Move Cost Couts de migration vers les voisins (tableau de taille nb_machines) */
};
typedef struct machine machine_t;


typedef struct {
	unsigned short prem_indice;   /* -1 si liste vide */
	unsigned short nbMaxelt;
	unsigned short *tab; 	   /* si non égal à -1 : indice sur elt suivant */
} liste_t;

typedef struct {
	unsigned short *tableauIndex;
	unsigned short *tableauChain;
} listeCompacte_t;

typedef struct {
	unsigned short **tableauIndex;
	unsigned short *tableauChain;
} listeCompacte2D_t;

struct service {
    unsigned  short spreadMin;         /* SpreadMin du service */
    unsigned  short nb_dep;            /* Nb de services dont le service depend */
    unsigned  short * dependances;      /* Autres services dont le service depend (tableau de taille nb_dep) */
    //liste_t ens_procs;    	/* liste des ps du service */
    unsigned  short nb_procs;
    unsigned  short nb_succ;           /* nbre de successeurs d'un service*/
    unsigned  short * successeurs;	/* successeurs dans la relation de dependance */
};
typedef struct service service_t;

struct balance_objective {
    unsigned short r1 ;              /* num. de ressource 1 */
    unsigned short r2;
    unsigned target;		  /* poids d'equilibrage */
    unsigned weight_balanceCost;
};
typedef struct balance_objective balance_objective_t;

struct instance { 
    unsigned short nb_neighborhood;       /* nbre de voisinages, voir dependances entre services */
    unsigned short nb_ressources;         /* Nombre de ressources des machines */
    unsigned short nb_transcient;	  /* Nbre de ressouces transcientes */
    ressource_t ressources[MAXNBRESSOURCES];   /* Capacites des differentes ressources des machines */
    unsigned short nb_machines;           /* Nombre de machines */
    machine_t *machines;       	    /* Machines : tableau de taille nb_machines */
    unsigned short nb_services;           /* Nombre de services */
    service_t *services;            /* Services : tableau de taille nb_services */
    listeCompacte_t services_ens_procs; /* liste des processus d'un service */
    unsigned short nb_procs;              /* Nombre de processus */
    processus_t* procs;        /* Processus : tableau de taille nb_procs */
    unsigned short nb_objectifs;          /* Nombre d'objectifs */
    balance_objective_t objectifs[MAXNBBALANCECOSTS]; /* Objectifs : tableau de taille nb_objectifs */
    unsigned short nb_locations;          /* nbre de localisation de machines */
    unsigned weight_PMC;            /* poids : process move cost */
    unsigned weight_SMC;            /* poids : service move cost */
    unsigned weight_MMC;            /* poids : machine move cost */
};
typedef struct instance instance_t;


struct placement{
#define NOMACHINE 	((unsigned short)-1u)
#define NOPROCESS 	((unsigned short)-1u)
#define NOSERVICE 	((unsigned short)-1u)
#define NOINIT 		((unsigned short)-2u)
    unsigned short *assignment; 	/* quel processus va sur quelle machine */
    unsigned **U;               /* U[m][r] : Charges actuelles des ressources de la machine */
    unsigned **residu;          /* residu[m][r] : charge qui reste en m après qu'un ps ait bougé, nécessaire pour transcient*/
    listeCompacte_t liste_procs;      /* liste_procs[m] : liste des n° de Processus assignes a la machine m*/
    unsigned short *nb_procs_par_mach;     /* nb_procs[m] : Nb de procs sur la machine m*/
    unsigned short **spread; 		/* spread[s][l]: nbre de ps du service s dans la location l */
    unsigned short *cardServiceMove; 	/* nbre de processus qui ont bougé par rapport à l'initial par service */
    unsigned short maxCardServiceMove;	/* le max sur le tableau précédent */
    listeCompacte2D_t liste_de_processus_d_un_service_dans_un_voisinage; /* liste de processus par service par n° de voisinage, indice d'entree du tab suivant */
    unsigned short **service_par_machine; 	/* = NOPROCESS si pas de ps de ce service, sinon LE ps de ce service */
#ifdef TRACE_ALGOS
    char * liste_algos;
    unsigned int nb_algos;
#endif
    int64  cout ;
};
typedef struct placement placement_t;

/* Variables globales */

/* Puisque instance contient les constantes du pb, on le définit comme variable globale 
 * pour gagner du temps dans les appels de fonction. Elle est réellement déclarée dans le main.*/
extern instance_t instance;

/* Pour la minuterie, attention doit etre de type sig_atomic_t */
extern volatile sig_atomic_t alarme;

/* Constantes */

#define FMR_OK          0
#define FMR_IO_ERROR   -1
#define FMR_NOT_FOUND  -2
#define FMR_OVERLOAD   -3
#define FMR_TIMEOUT    -4
#define FMR_NULL_PTR   -5

#define DERNIERE_MIN	20

/* Macros */

#define MAX( a, b ) (a) > (b) ? (a) : (b) 
#define MIN( a, b ) (a) < (b) ? (a) : (b) 
#ifdef __GNUC__
#define LIKELY(x)       __builtin_expect((x),1)
#define UNLIKELY(x)     __builtin_expect((x),0)
#else
#define LIKELY(x)       (x)
#define UNLIKELY(x)     (x)
#endif


#ifdef DEBUG
# define DEBUG_PRINT(...) printf(__VA_ARGS__)
#else
# define DEBUG_PRINT(...) 
#endif

#ifdef DEBUGIO
# define DEBUG_IO_PRINT(...) printf(__VA_ARGS__)
#else
# define DEBUG_IO_PRINT(...) 
#endif

#ifdef DEBUGSTATS
# define DEBUG_STATS_PRINT(...) printf(__VA_ARGS__)
#else
# define DEBUG_STATS_PRINT(...) 
#endif

#endif
