/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "fmr.h"
#include "fmr_tools.h"

#include "fmr_constraints.h"

bool verifU(placement_t *placmnt, unsigned int m) {
	unsigned U;
	unsigned int p,r;

	for (r=0; r<instance.nb_ressources; r++) {
		U=0;
		for (p=0;p<instance.nb_procs; p++) {
	   	   if (placmnt->assignment[p] == m)
	    		 U += instance.procs[p].R[r];
		}
		if  (U!= placmnt->U[m][r]) {
			printf("erreur pour r=%d\n",r);
			return false;
		}
	}
	return(true);
}

/* vérifie que la suppression du ps p de la machine m, laisse une solution valide */
bool
check_remove_proc (unsigned int p, unsigned int m, placement_t * placmnt)
{
    unsigned int s = instance.procs[p].service;
    unsigned int n = instance.machines[m].neighborhood;
    unsigned int l = instance.machines[m].location;
    unsigned int ll;
    unsigned int nblocadiff;
    liste_t *ptr;

/* rien à faire pour la contrainte de capacité */
/* ni pour Conflicts constaints */

/* voisinage: dependancy constraint */
    //ptr = & placmnt->liste_de_processus_d_un_service_dans_un_voisinage[n][s];
    if (instance.services[s].nb_succ != 0) {// ! liste_est_vide(&instance.serviceSuccesseur[s])) {
	//if (liste_est_singleton (ptr))	/* zut ils vont devenir orphelins ! */
	if (listeCompacte2D_est_singleton(& placmnt->liste_de_processus_d_un_service_dans_un_voisinage,n,s))
	    return (false);
    }

/* Spread constraint : les ps d'un service doivent s'étaler sur au min spreadMin locations diff. */
    if (placmnt->spread[s][l] == 1) {	/* donc risque que nb de location diff diminue */
	nblocadiff = 0;
	for (ll = 0; ll < instance.nb_locations; ll++) {
	    if (ll != l && (placmnt->spread[s][ll] > 0))
		nblocadiff++;
	}
	if (nblocadiff < instance.services[s].spreadMin)
	    return (false);
    }
    return (true);
}


/* vérifie que l'ajout du processus p sur la machine m laisse une solution valide */
bool
check_add_proc (const placement_t const *initial, unsigned int p, unsigned int m,
		placement_t * placmnt)
{
    unsigned int i, r, sb;
    unsigned int m0 = initial->assignment[p];
    unsigned int s = instance.procs[p].service;
    unsigned int n_dest = instance.machines[m].neighborhood;
    liste_t *ptr;

    /* load constraint + transcient */
    for (r = 0; r < instance.nb_ressources; r++) {
	if (m != m0 && instance.ressources[r].transcient) {
	    if ((unsigned long) placmnt->U[m][r] + instance.procs[p].R[r] + placmnt->residu[m][r] > instance.machines[m].C[r]) {
		//DEBUG_PRINT("check_add_proc transcient: load constraint p=%d, m=%d, r=%d, U:%d, p:%d, residu:%d, C=%d\n", p, m, r, placmnt->U[m][r], instance.procs[p].R[r],
		//		placmnt->residu[m][r], instance.machines[m].C[r]) ;
		return (false);
	    }
    } else if ((unsigned long) placmnt->U[m][r] + instance.procs[p].R[r] + placmnt->residu[m][r] > instance.machines[m].C[r]) {
		DEBUG_PRINT("check_add_proc: load constraint p=%d,  m=%d, r=%d, U:%d, p:%d,  C=%d\n", p, m, r, placmnt->U[m][r], instance.procs[p].R[r],
			 instance.machines[m].C[r]) ;
	    return (false);
	}
    }


    /* conflict constraint */
    if (placmnt->service_par_machine[m][s] != NOPROCESS) {	/* il y a déjà un ps du service s sur m_dest */
	return (false);
    }

    /* pas de pb avec contrainte spread: on ne peut qu'augmenter le card. en ajoutant un ps */

    /* depandancy constraints */
    /* vérification des prédécesseurs de p */
    for (i = 0; i < instance.services[s].nb_dep; i++) {
	sb = instance.services[s].dependances[i];
	//ptr = & placmnt->liste_de_processus_d_un_service_dans_un_voisinage[n_dest][sb];
	//if (liste_est_vide(ptr))	/* existe -il au moins un processus du service sb appartenant au voisinage n_dest */
	if (listeCompacte2D_est_vide(& placmnt->liste_de_processus_d_un_service_dans_un_voisinage,n_dest,sb))
	    return (false);
    }

    return (true);
}

/* Capacity constraint, vraie si contraintes de capacitées sont vérifiées  */
bool capacity_constraint( placement_t* placmnt ){
	unsigned int r,m,p;

	/* Vérification de U(m,r) <= C(m,r) */
	for (m=0; m< instance.nb_machines; m++) {
		for (r=0; r<instance.nb_ressources; r++) {
			/* double vérif !  A commenter si tout va bien */
			unsigned int U=0;
			for (p=0;p<instance.nb_procs; p++) {
				if (placmnt->assignment[p] == m)
					U += instance.procs[p].R[r];
			}
			if (U != placmnt->U[m][r]) {
				DEBUG_PRINT("charge de %d(r=%d) fausse (U= %d, U calculé= %d)\n",m,r,placmnt->U[m][r],U);
			}
			/* Fin de zone double vérif */

			if (placmnt->U[m][r]+ placmnt->residu[m][r] > instance.machines[m].C[r]) {
				DEBUG_PRINT("charge de %d trop grande (%d/%d)\n", m, placmnt->U[m][r], instance.machines[m].C[r]);
				return(false);
			}
		}
	}
	return(true);
}

/*1 seul déplact de processus */
bool capacity_constraint_shift ( placement_t* placmnt, unsigned int p, unsigned int m_orig, unsigned int m_dest) {
	unsigned int r;

	for (r=0; r<instance.nb_ressources; r++) {
        if ((unsigned long)placmnt->U[m_dest][r] + instance.procs[p].R[r] +placmnt->residu[m_dest][r]> instance.machines[m_dest].C[r])
			return(false);
	}
	return(true);
}

/* cas de l'échange de procs */
bool capacity_constraint_swap (placement_t* placmnt, unsigned int p1, unsigned int m1, unsigned int p2, unsigned int m2) {
	unsigned int r;

	for (r=0; r<instance.nb_ressources; r++) {
	   if (placmnt->U[m1][r] +instance.procs[p2].R[r] +placmnt->residu[m1][r]> instance.machines[m1].C[r]
		|| placmnt->U[m2][r] +instance.procs[p1].R[r] +placmnt->residu[m2][r]>instance.machines[m2].C[r])
		return(false);
	}
	return(true);
}

/* Conflicts constaints : pas 2 ps d'un meme service sur une meme machine*/
bool conflicts_constraint ( placement_t* placmnt) {
	unsigned short s;
	unsigned short e1,e2;
	//liste_t *l;

	for (s=0;s<instance.nb_services;s++) {
		//l = &instance.services[s].ens_procs;
		e1=listeCompacte_tete(& instance.services_ens_procs, s);
		//e1= liste_head(l);
		while ( e1 != LISTE_VIDE_US) {
			e2=listeCompacte_next(& instance.services_ens_procs, s, e1); /* e2 <> e1 */
			while ( e2 != LISTE_VIDE_US ) {
				if (placmnt->assignment[e1] == placmnt->assignment[e2]) {
					DEBUG_PRINT("Attention: %d et %d sur la même machine %d!\n",e1, e2,placmnt->assignment[e1]);
					return(false);
				}
				e2 = listeCompacte_next(& instance.services_ens_procs, s, e2);
			}
			e1 = listeCompacte_next(& instance.services_ens_procs, s, e1);
		}
	}
	return(true);
}

/* 1 seul deplt de processus */
bool conflicts_constraint_shift ( placement_t* placmnt, unsigned int p, unsigned int m_dest) {
	unsigned int s= instance.procs[p].service;
	/*liste_t *e1 = instance.services[s].ens_procs;

	while ( e1 != NULL) {
		if (e1->id != p && placmnt->assignment[e1->id]== m_dest)
			return(false);
		e1=e1->next;
	}*/
	if (placmnt->service_par_machine[m_dest][s] != NOPROCESS) {/* il y a déjà un ps du service s sur m_dest*/
		return(false);
	}

	return(true);
}

/* cas de l'échange de ps : on suppose m1!=m2 */
bool conflicts_constraint_swap ( placement_t* placmnt, unsigned int p1, unsigned int m1, unsigned int p2, unsigned int m2) {
	unsigned int s1 = instance.procs[p1].service;
	unsigned int s2 = instance.procs[p2].service;

	if (s1==s2) /*sachant que m1!= m2*/
	  return(true);

	if (placmnt->service_par_machine[m2][s1] != NOPROCESS || placmnt->service_par_machine[m1][s2] != NOPROCESS)
	  return(false);
	return(true);
}

/* Spread constraint : les ps d'un service doivent s'étaler sur au min spreadMin locations diff. */
bool spread_constraint (placement_t *placmnt) {
	unsigned int s,l,nblocadiff;
	/* spread[s][l] donne le nbre de processus du service s courant placés sur des machines de l */

	for (s=0; s< instance.nb_services; s++) {
		nblocadiff=0;
		for (l=0;l<instance.nb_locations;l++) {
			if (placmnt->spread[s][l]>0)
				nblocadiff++;
		}
		if (nblocadiff< instance.services[s].spreadMin) {
			DEBUG_PRINT("spread violée pour s=%d (%d au lieu de %d)\n",s, nblocadiff, instance.services[s].spreadMin);
			return(false);
		}
	}
	return(true);
}

/* 1 seul mvt */
bool spread_constraint_shift (placement_t *placmnt, unsigned int p, unsigned int m_orig, unsigned int m_dest) {
	unsigned int s, l, l_orig, l_dest ;
	unsigned int nblocadiff;

	s=instance.procs[p].service;
	l_orig =instance.machines[m_orig].location;
	l_dest=instance.machines[m_dest].location;
	if (l_orig == l_dest /* Pas de chgt */
		|| placmnt->spread[s][l_orig ] > 1 /* donc le nb de location diff ne diminuera pas */
		|| placmnt->spread[s][l_dest] == 0) /* idem */
			return(true);

	/* reste les autres cas: placmnt->spread[s][l_orig ]==1 et placmnt->spread[s][l_dest]>=1
	 * il faut encore vérifier que la diminution de 1 ne va pas faire passer sous le seuil */
	nblocadiff=0;
	for (l=0;l<instance.nb_locations;l++) {
		if (l!=l_orig && (placmnt->spread[s][l]>0))
			nblocadiff++;
	}
	return(nblocadiff >= instance.services[s].spreadMin);
}


/* cas de l'échange de ps : on suppose m1!=m2 */
bool spread_constraint_swap (placement_t *placmnt, unsigned int p1, unsigned int m1, unsigned int p2, unsigned int m2) {
	unsigned int s1=instance.procs[p1].service;
	unsigned int s2=instance.procs[p2].service;
	unsigned int l1 =instance.machines[m1].location;
	unsigned int l2 =instance.machines[m2].location;

	if (l1 == l2) /* Pas de chgt sur le nbre de locations diff.*/
	  return(true);
	if (s1 == s2) /* idem */
	   return(true);

  	/* cela revient à 2 deplt indépendants */
	return( spread_constraint_shift(placmnt,p1,m1,m2) && spread_constraint_shift(placmnt,p2, m2,m1));
}

/* Dependancy constraints */
bool dependancy_constraint (placement_t *placmnt) {
	unsigned int n,ma,mb,sa,sb,i;
	//liste_t *pa, *pb;
	unsigned short ppa, ppb;
	bool trouve;

	for (sa=0; sa< instance.nb_services; sa++) {
		if (instance.services[sa].nb_dep>0) {
			//pa = & instance.services[sa].ens_procs;
			//ppa = liste_head(pa);
			ppa = listeCompacte_tete(& instance.services_ens_procs, sa);

			while (ppa != LISTE_VIDE_US) {
				ma= placmnt->assignment[ppa];
				n= instance.machines[ma].neighborhood;
			      	for (i=0; i< instance.services[sa].nb_dep;i++) {
					sb= instance.services[sa].dependances[i];
					//pb =& instance.services[sb].ens_procs;
					//ppb = liste_head(pb);
					ppb = listeCompacte_tete(& instance.services_ens_procs, sb);
					trouve=false;
					while (ppb != LISTE_VIDE_US && ! trouve) {
						mb=placmnt->assignment[ppb] ;
						trouve = (instance.machines[mb].neighborhood == n);
						ppb = listeCompacte_next(& instance.services_ens_procs, sb, ppb);
					}
					if (!trouve)
					{
						DEBUG_PRINT("dependancy violée pour serv_a=%d, proc_a=%d sur mach_a=%d, neighb=%d depend_serv_b=%d\n",sa, ppa, ma, n, sb);
						return(false);	
					}
			    }
			    ppa = listeCompacte_next(& instance.services_ens_procs, sa, ppa);
			}
		}
	}
	return(true);
}

/* 1 seul mvt */
bool dependancy_constraint_diff (placement_t *placmnt, unsigned int pa, unsigned int m_orig, unsigned int m_dest) {
	unsigned int n_orig= instance.machines[m_orig].neighborhood;
	unsigned int n_dest= instance.machines[m_dest].neighborhood;
	unsigned int sb, sa = instance.procs[pa].service,  i;
	liste_t *ptr;

	if (n_orig == n_dest ) /* pas de changement */
		return(true);

	/* vérification des prédécesseurs de pa */
	for (i=0; i< instance.services[sa].nb_dep;i++) {
                sb = instance.services[sa].dependances[i];
		if (listeCompacte2D_est_vide(& placmnt->liste_de_processus_d_un_service_dans_un_voisinage,n_dest,sb))
			return(false);
	}

	/* attention à ne pas oublier les successeurs du service sa */
	// y a-t-il un service dépendant de sa dans le voisinage n_orig ?
	if (listeCompacte2D_est_singleton(& placmnt->liste_de_processus_d_un_service_dans_un_voisinage, n_orig, sa)) {
	/*	return(false);*/
	    for (i=0;i< instance.services[sa].nb_succ;i++) {
		sb=instance.services[sa].successeurs[i];
		if (! listeCompacte2D_est_vide (&placmnt->liste_de_processus_d_un_service_dans_un_voisinage, n_orig, sb))
		   return false ;
	    }
	}
	return(true);
}

/* cas de l'échange de ps : on suppose m1!=m2 */
bool dependancy_constraint_swap (placement_t *placmnt, unsigned int p1, unsigned int m1, unsigned int p2, unsigned int m2) {
	unsigned int n1 =instance.machines[m1].neighborhood;
	unsigned int n2 =instance.machines[m2].neighborhood;
	unsigned int s1 =instance.procs[p1].service;
	unsigned int s2 =instance.procs[p2].service;

	if (n1 == n2) /* pas de changement de voisinage */
		return(true);
	if (s1 == s2) /* l'un compense l'autre */
		return(true);

	/*  2 deplct independants (hum j'ai un doute) */
	return(dependancy_constraint_diff(placmnt, p1, m1, m2) && dependancy_constraint_diff(placmnt, p2, m2, m1));
}

/* Transient usage constraints */
bool transient_constraint (const placement_t const *initial, placement_t *resu) {
	unsigned int r,m,som,p;

	for (m=0; m<instance.nb_machines;m++) {
		for (r=0; r<instance.nb_ressources; r++) {
			if (instance.ressources[r].transcient) {
				som=0;
				for (p=0;p<instance.nb_procs;p++) {
					if (initial->assignment[p]==m || (resu->assignment[p]==m))
						som += instance.procs[p].R[r];
				}
				if (som > instance.machines[m].C[r]) {
					DEBUG_PRINT("Transcient violée par m=%d, r=%d\n, (%d/%d)",m,r, som, instance.machines[m].C[r]);
					return(false);
				}

			}
		}
	}
	return(true);
}

/* 1 seul mvt */
bool transient_constraint_shift (placement_t *initial, placement_t *placmnt, unsigned int proc, unsigned int m_orig, unsigned int m_dest) {
//	int m0= initial->assignment[proc];
 /* contrainte déjà réglée par la containte de capacité :
	soit proc va de m0 à une autre, soit d'une autre vers m0 soit entre deux diff. de m0*/
	return(true);
}

bool check_all_constraints (const placement_t const *initial, placement_t *resu) {
	// Inutile normalement car déjà vérifier dans l'explo.
	if (! capacity_constraint( resu )) {
		DEBUG_PRINT("erreur capacity_constraint\n");
		return(false);
	}
	if (! conflicts_constraint ( resu)) {
		DEBUG_PRINT("erreur conflicts_constraint\n");
		return(false);
	}
	if (! spread_constraint (resu)) {
		DEBUG_PRINT("erreur spread_constraint\n");
		return(false);
	}
	if (! dependancy_constraint (resu)) {
		DEBUG_PRINT("erreur dependancy_constaint\n");
		return(false);
	}
	if (! transient_constraint (initial, resu)) {
		DEBUG_PRINT("erreur transient_constaint\n");
		return(false);
	}
	return(true);

}

/* 1 seul mvt */
bool check_all_constraints_shift( placement_t *pl_temp, unsigned int p , unsigned int m_orig, unsigned int m_dest) {
	/* A FAIRE: mettre dans l'ordre du temps de calcul par contrainte et/ou de fréquence de violation? */
	if (! capacity_constraint_shift(pl_temp, p, m_orig, m_dest))
		return(false);
	if (! conflicts_constraint_shift(pl_temp, p, m_dest))
		return(false);
	if (! spread_constraint_shift (pl_temp,  p, m_orig,  m_dest) )
		return(false);
	if (! dependancy_constraint_diff (pl_temp, p, m_orig, m_dest))
		return(false);
//	if (! transient_constraint_diff (pl_temp,  p,  m_dest) )
//		return(false);
	return(true);
}

/* cas de l'échange de ps : on suppose m1!=m2 */
bool check_all_constraints_swap( placement_t *pl_temp, unsigned int p1, unsigned int m1, unsigned int p2, unsigned int m2) {
	/* A FAIRE: mettre dans l'ordre du temps de calcul par contrainte et/ou de fréquence de violation? */
	if (! capacity_constraint_swap(pl_temp, p1, m1, p2, m2))
		return(false);
	if (! conflicts_constraint_swap(pl_temp, p1, m1, p2, m2))
		return(false);
	if (! spread_constraint_swap(pl_temp, p1, m1, p2, m2))
		return(false);
	if (! dependancy_constraint_swap(pl_temp, p1, m1, p2, m2))
		return(false);
	return(true);
}

/* vérifie que l'ajout du processus p sur la machine m laisse une solution valide */
bool
check_add_proc_glouton (const placement_t const *initial, unsigned int p, unsigned int m, placement_t * placmnt)
{
    unsigned int i, r, sb;
    unsigned int m0 = initial->assignment[p];
    unsigned int s = instance.procs[p].service;
    unsigned int n_dest = instance.machines[m].neighborhood;
    liste_t *ptr;

    /* load constraint + transcient */
    for (r = 0; r < instance.nb_ressources; r++) {
        if (m != m0 && instance.ressources[r].transcient) {
            if ((unsigned long) placmnt->U[m0][r] + instance.procs[p].R[r] + placmnt->residu[m0][r] > instance.machines[m0].C[r]

                || (unsigned long) placmnt->U[m][r] + instance.procs[p].R[r] + placmnt->residu[m][r] > instance.machines[m].C[r]) {
                return false;
            }
        } else if ((unsigned long) placmnt->U[m][r] + instance.procs[p].R[r] + placmnt->residu[m][r] > instance.machines[m].C[r]) {
            return false;
        }
    }

    /* conflict constraint */
    if (placmnt->service_par_machine[m][s] != NOPROCESS) {     /* il y a déjà un ps du service s sur m_dest */
	    DEBUG_PRINT("Il y a déjà le ps %d de service %d sur %d\n",placmnt->service_par_machine[m][s],s,m);
        return false;
    }

    /* pas de pb avec contrainte spread: on ne peut qu'augmenter le card. en ajoutant un ps */

    /* depandancy constraints */
    /* vérification des prédécesseurs de p */
    for (i = 0; i < instance.services[s].nb_dep; i++) {
	sb = instance.services[s].dependances[i];
	//ptr = & placmnt->liste_de_processus_d_un_service_dans_un_voisinage[n_dest][sb];
	//if (liste_est_vide(ptr))	/* existe -il au moins un processus du service sb appartenant au voisinage n_dest */
	if (listeCompacte2D_est_vide( & placmnt->liste_de_processus_d_un_service_dans_un_voisinage, n_dest, sb))
	    return (false);
    }
    /* on ne vérifie pas tout de suite les dépendances par voisinage */
     return true;
}


/* vérifie que l'ajout du processus p sur la machine m laisse une solution valide */
bool
check_add_proc_glouton2 (const placement_t const *initial, unsigned int p, unsigned int m, placement_t * placmnt)
{
    unsigned int r;
    unsigned int m0 = initial->assignment[p];

    /* load constraint + transcient */
    for (r = 0; r < instance.nb_ressources; r++) {
        if (m != m0) {
            if  (instance.ressources[r].transcient && (unsigned long) instance.procs[p].R[r] + initial->U[m][r] > instance.machines[m].C[r])
                   return false;
        }
    }

    /* pas de pb avec contrainte spread: on ne peut qu'augmenter le card. en ajoutant un ps */

    /* on ne vérifie pas tout de suite les dépendances par voisinage */
     return true;
}

