/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/

#include <stdio.h>
#include <stdint.h>

#include "fmr.h"
#include "fmr_cost.h"


long A(placement_t *placmt, unsigned int m, unsigned int r) {
	return ((signed long) instance.machines[m].C[r] - (signed long)placmt->U[m][r]);
}
static inline int max(int a, int b) {
    return (a>b ? a : b);
}

/* calcule le gain réalisé si deplact d'1 proc */
int64_t
calcule_gain_shift (const placement_t const *initial, placement_t *placmt, unsigned short p, unsigned short m_orig, unsigned short m_dest) {
	unsigned int target;
	uint64_t balance_cost_orig_avant, balance_cost_orig_apres, balance_cost_dest_avant, balance_cost_dest_apres;
	unsigned short i, b,r,r1,r2, newmax;
	unsigned short m0= initial->assignment[p];
	unsigned short s=instance.procs[p].service;
   	int64 totalGain=0, totalGainR;
	unsigned long seuil_orig;
    int64 seuil_dest;

	/* loadCost*/
	for (r=0; r< instance.nb_ressources;r++) {
	    totalGainR=0;
	    seuil_orig =instance.procs[p].R[r] +instance.machines[m_orig].SC[r];
	    seuil_dest= - (long)instance.procs[p].R[r] + (long)instance.machines[m_dest].SC[r];
	    if ( placmt->U[m_orig][r] >= seuil_orig) { /* donc gain sur m_orig*/
	  	totalGainR = (long) instance.procs[p].R[r];
	    } else  if (  placmt->U[m_orig][r] > instance.machines[m_orig].SC[r]) { /* donc gain pas nul sur m_orig */
	        totalGainR = (long)placmt->U[m_orig][r] - (long)instance.machines[m_orig].SC[r];
	    }
	    /* else gain nul sur m_orig */
	    if (placmt->U[m_dest][r] >=  instance.machines[m_dest].SC[r]) { /* donc perte sur m dest */
		totalGainR -= (long)instance.procs[p].R[r];
	    } else  if ( (signed) placmt->U[m_dest][r] > seuil_dest) {
		totalGainR -= (long)placmt->U[m_dest][r] + (long)instance.procs[p].R[r] - (long)instance.machines[m_dest].SC[r];
	    }
	    /* else perte nulle sur m dest */

            totalGain += totalGainR * (long)instance.ressources[r].weight_loadCost;
	}
	//DEBUG_PRINT("gain_shift(%d de %d vers %d): loadcost=%ld,",p,m_orig,m_dest,totalGain);

	/* Machine Move Cost*/
	if (m0 == m_dest) { /* deplact vers M0 donc diminution du cout */
		totalGain += (long)(instance.weight_MMC * instance.machines[m0].MMC[m_orig]);
	} else if (m0 == m_orig) {/* deplact depuis M0 donc augmentation du cout */
		totalGain -= (long)(instance.weight_MMC * instance.machines[m0].MMC[m_dest]);
	} else /* on gagne du coté orig on perd du coté dest */
	   	totalGain -= (long)instance.weight_MMC * ((long)instance.machines[m0].MMC[m_dest]- (long)instance.machines[m0].MMC[m_orig]);

	//DEBUG_PRINT("%ld,",totalGain);
	/* Balance cost */
    	for( b = 0 ; b < instance.nb_objectifs ; b++ ) {
       	 	r1 = instance.objectifs[b].r1;
        	r2 = instance.objectifs[b].r2;

      		target = instance.objectifs[b].target;

		long int temp = ((long int) target) * A(placmt,m_orig,r1) - (A(placmt,m_orig,r2));
		balance_cost_orig_avant = (temp<0 ? 0 : (unsigned long) temp);

		temp =  (long)target * (A(placmt,m_orig,r1)+(long)instance.procs[p].R[r1]) - (A(placmt,m_orig,r2)+(long)instance.procs[p].R[r2]);
		balance_cost_orig_apres = (temp<0 ? 0 : (unsigned long) temp);

	        temp = (long)target * A(placmt,m_dest,r1) - A(placmt,m_dest,r2);
		balance_cost_dest_avant=(temp<0 ? 0 : (unsigned long) temp);

	        temp = (long)target * (A(placmt,m_dest,r1)-(long)instance.procs[p].R[r1]) - (A(placmt,m_dest,r2)-(long)instance.procs[p].R[r2]);
		balance_cost_dest_apres=(temp<0 ? 0 : (unsigned long) temp);

                totalGain -= (signed)instance.objectifs[b].weight_balanceCost *
			((signed)balance_cost_orig_apres-(signed)balance_cost_orig_avant + (signed)balance_cost_dest_apres-(signed)balance_cost_dest_avant);
        }
	//DEBUG_PRINT("%ld,",totalGain);

	/* Process move cost */
	if (m_orig == m0) {
		totalGain -= (long)(instance.weight_PMC * instance.procs[p].PMC);
	} else if (m_dest == m0) {
		totalGain += (long)(instance.weight_PMC * instance.procs[p].PMC);
	}
	/* else pas de changement */
	//DEBUG_PRINT("%ld,",totalGain);


	/* Service move cost */
	if (m_orig == m0) { /* il a bougé de m0 : il faut payer */
		if (placmt->maxCardServiceMove == placmt->cardServiceMove[s] ) /* le max va augmenter de 1*/
			totalGain -= (long)instance.weight_SMC ;
		/* else le max ne change pas donc pas de changement de cout */
	} else if (m_dest == m0) { /* il retourne vers m0 : gain potentiel */
		if (placmt->maxCardServiceMove == placmt->cardServiceMove[s] ) {/* le max peut éventuellement changer */
			newmax=0;
			for (i=0; i< instance.nb_services; i++) {
				if (i==s) {
				  if (newmax < placmt->cardServiceMove[i] -1)
					newmax = (unsigned short)(placmt->cardServiceMove[i] -1);
				}
				else if (newmax < placmt->cardServiceMove[i])
					newmax = placmt->cardServiceMove[i];
			}
			if (newmax < placmt->maxCardServiceMove) /* le max va diminuer de 1 */
				totalGain += (long)instance.weight_SMC;
		}
	}
	//DEBUG_PRINT("%ld\n",totalGain);
	/* else pas de changement */
	return (totalGain);
}

int64_t
calcule_gain_ajout(const placement_t *initial, placement_t *placmt, unsigned short p, unsigned short m_dest) {
        unsigned int target;
        long avant, apres;
        unsigned short b, r,r1,r2, m0= initial->assignment[p];
        unsigned short s=instance.procs[p].service;
        long totalGain=0;

        /* loadCost*/
        for (r=0; r< instance.nb_ressources;r++) {
           if  ( (avant=(long) placmt->U[m_dest][r]-(long)instance.machines[m_dest].SC[r]) <0)
                    avant=0;
            if ( (apres=(long) placmt->U[m_dest][r]+(long)instance.procs[p].R[r] - (long)instance.machines[m_dest].SC[r]) <0)
                    apres=0;
            totalGain -= (apres-avant)* (long)instance.ressources[r].weight_loadCost;
	}

	//DEBUG_PRINT("gain_ajout: loadcost=%ld,",totalGain);
       /* Balance cost */
        for( b = 0 ; b < instance.nb_objectifs ; b++ ) {
                r1 = instance.objectifs[b].r1;
                r2 = instance.objectifs[b].r2;

                target = instance.objectifs[b].target;

                avant = (long)target * A(placmt,m_dest,r1) - A(placmt,m_dest,r2);
                if (avant <0)
                        avant=0;
                apres = (long)target * (A(placmt,m_dest,r1)-(long)instance.procs[p].R[r1])
                        - (A(placmt,m_dest,r2)-(long)instance.procs[p].R[r2]);
                if (apres <0)
                        apres=0;

                totalGain -= (long)instance.objectifs[b].weight_balanceCost * (apres-avant);
        }
	//DEBUG_PRINT("%ld,",totalGain);

        if (m_dest != m0) {
        	/* Machine Move Cost*/
                totalGain -= (long)(instance.weight_MMC * instance.machines[m0].MMC[m_dest]);
        	/* Process move cost */
                totalGain -= (long)(instance.weight_PMC * instance.procs[p].PMC);
        	/* Service move cost */
                if (placmt->maxCardServiceMove == placmt->cardServiceMove[s] ) /* le max va augmenter de 1*/
                        totalGain -= (long)instance.weight_SMC ;
        }
	//DEBUG_PRINT("%ld\n",totalGain);
        return (totalGain);
}


/* calcule le gain réalisé si suppresion d'1 proc */
long
calcule_gain_suppr (const placement_t const *initial, placement_t *placmt, unsigned short p, unsigned short m_orig) {
	unsigned int target;
	long avant, apres;
	unsigned short i,b,r,r1,r2,newmax;
	unsigned short m0= initial->assignment[p];
	unsigned short s=instance.procs[p].service;
   	long totalGain=0;

	/* loadCost*/
	for (r=0; r< instance.nb_ressources;r++) {
	    if  ( (avant=(long) placmt->U[m_orig][r]-(long)instance.machines[m_orig].SC[r]) <0)
		    avant=0;
        if ( (apres=(long) placmt->U[m_orig][r]-(long)instance.procs[p].R[r] -(long)instance.machines[m_orig].SC[r]) <0)
		    apres=0;
        totalGain -= (apres-avant)* (long)instance.ressources[r].weight_loadCost;
	}

	if (m0 != m_orig) { /* p n'était pas sur sa m0 donc gain */
		/* Machine Move Cost*/
        totalGain += (long)instance.weight_MMC * (long)instance.machines[m0].MMC[m_orig];
		/* Process move cost */
        totalGain += (long)instance.weight_PMC * (long)instance.procs[p].PMC;
		/* Service move cost */
		if (placmt->maxCardServiceMove == placmt->cardServiceMove[s] ) {/* le max peut éventuellement diminuer*/
			newmax=0;
			for (i=0; i< instance.nb_services; i++) {
				if (i==s && newmax < placmt->cardServiceMove[i] -1)
					newmax = (unsigned short)(placmt->cardServiceMove[i] -1);
				else if (newmax < placmt->cardServiceMove[i])
					newmax = placmt->cardServiceMove[i];
			}
			if (newmax < placmt->maxCardServiceMove) /* le max va diminuer de 1 */
                totalGain += (long)instance.weight_SMC;
		}
	}

	/* Balance cost */
    	for( b = 0 ; b < instance.nb_objectifs ; b++ ) {
       	 	r1 = instance.objectifs[b].r1;
        	r2 = instance.objectifs[b].r2;

      		target = instance.objectifs[b].target;
		avant = (long)target * A(placmt,m_orig,r1) - A(placmt,m_orig,r2);
		if (avant <0)
			avant=0;
        apres = (long)target * (A(placmt,m_orig,r1)+(long)instance.procs[p].R[r1])
        - (A(placmt,m_orig,r2)+(long)instance.procs[p].R[r2]);
		if (apres <0)
			apres=0;

        totalGain -= (long)instance.objectifs[b].weight_balanceCost * (apres-avant);
        }

	return (totalGain);
}

/* calcule le gain réalisé si echange de ps de machines diff. sans faire effectivement l'échange*/
long
calcule_gain_swap (const placement_t const *initial, placement_t *placmt, unsigned short p1, unsigned short m1, unsigned short p2, unsigned short m2) {
	unsigned short i, b, r,r1,r2;
	unsigned int  target;
	unsigned short newmax;
	unsigned short m01= initial->assignment[p1];
	unsigned short m02= initial->assignment[p2];
	unsigned short s1=instance.procs[p1].service;
	unsigned short s2=instance.procs[p2].service;
   	long totalGain=0, avant, apres;
	bool recalculMax;

	/* loadCost*/
	for (r=0; r< instance.nb_ressources;r++) {
	    if  ( (avant=(long) placmt->U[m1][r]-(long)instance.machines[m1].SC[r]) <0)
		    avant=0;
	    if ( (apres=(long) placmt->U[m1][r]-(long)instance.procs[p1].R[r]+(long)instance.procs[p2].R[r] -(long)instance.machines[m1].SC[r]) <0)
		    apres=0;
        totalGain -= (apres-avant)* (long)instance.ressources[r].weight_loadCost;
        if  ( (avant=(long) placmt->U[m2][r]-(long)instance.machines[m2].SC[r]) <0)
		    avant=0;
        if ( (apres=(long) placmt->U[m2][r]+(long)instance.procs[p1].R[r]-(long)instance.procs[p2].R[r] -(long)instance.machines[m2].SC[r]) <0)
		    apres=0;
        totalGain -= (apres-avant)* (long)instance.ressources[r].weight_loadCost;
	}
//	DEBUG_PRINT("gainswap %d <->%d Load cost=%ld,",p1,p2,totalGain);

	/* Machine Move Cost*/
	avant=0;
	if (m01!=m1)
        avant = (long)instance.machines[m01].MMC[m1];
	apres=0;
	if (m2!=m01)
        apres = (long)instance.machines[m01].MMC[m2];
    totalGain -= (apres-avant) * (long)instance.weight_MMC;
	avant=0;
	if (m02!=m2)
        avant = (long)instance.machines[m02].MMC[m2];
	apres=0;
	if (m1!=m02)
        apres = (long)instance.machines[m02].MMC[m1];
    totalGain -= (apres-avant) * (long)instance.weight_MMC;

	//DEBUG_PRINT("MMC=%ld,",totalGain);


	/* Balance cost */
    	for( b = 0 ; b < instance.nb_objectifs ; b++ ) {
       	 	r1 = instance.objectifs[b].r1;
        	r2 = instance.objectifs[b].r2;
      		target = instance.objectifs[b].target;

		/* du coté de m1 */
		avant = (long)target * A(placmt,m1,r1) - A(placmt,m1,r2);
		if (avant <0) avant=0;

			/* on ajoute p2 et on enleve p1*/
            apres = (long)target * (A(placmt,m1,r1)+(long)instance.procs[p1].R[r1]-(long)instance.procs[p2].R[r1])
            - (A(placmt,m1,r2)+(long)instance.procs[p1].R[r2]-(long)instance.procs[p2].R[r2]);
		if (apres <0) apres=0;
		totalGain -= (apres-avant)* (long)instance.objectifs[b].weight_balanceCost;

		/* du coté de m2 */
		avant = (long)target * A(placmt,m2,r1) - A(placmt,m2,r2);
		if (avant <0) avant=0;

apres = (long)target * (A(placmt,m2,r1)+(long)instance.procs[p2].R[r1]-(long)instance.procs[p1].R[r1])
- (A(placmt,m2,r2)+(long)instance.procs[p2].R[r2]-(long)instance.procs[p1].R[r2]);
		if (apres <0) apres=0;
		totalGain -= (apres-avant)* (long)instance.objectifs[b].weight_balanceCost;
        }
	//DEBUG_PRINT("balance=%ld,",totalGain);

	/* Process move cost */
	avant=0;
    if (m1!=m01) avant=(long)instance.procs[p1].PMC;
	apres=0;
    if (m2!=m01) apres=(long)instance.procs[p1].PMC;
    totalGain -= (apres-avant) * (long)instance.weight_PMC;

	avant=0;
    if (m2!=m02) avant=(long)instance.procs[p2].PMC;
	apres=0;
    if (m1!=m02) apres=(long)instance.procs[p2].PMC;
    totalGain -= (apres-avant) * (long)instance.weight_PMC;

	//DEBUG_PRINT("PMC=%ld,",totalGain);

	/* Service move cost */

    recalculMax=false;
    newmax=0;
    /* deltas potentiels des services s1 et s2 :
     * si la machine d'origine est la machine initiale : 1
     * si la machine de destination est la machine initiale : -1
     * sinon 0 (le déplacement du processus a déjà été compté */
    short int delta1=(m1==m01 ? 1 : m2==m01 ? -1 : 0);
    short int delta2=(m2==m02 ? 1 : m1==m02 ? -1 : 0);
    
    if (s1==s2 && placmt->cardServiceMove[s1] >= placmt->maxCardServiceMove-1) 
    { /* les processus sont sur le même service et peuvent modifier le maxCardServiceMove */
        int delta = delta1 + delta2;
        /* nombre de déplacements dans le service */
        newmax = (unsigned short) ((signed short)placmt->cardServiceMove[s1] + delta) ;
        if ((delta = (signed short)(newmax - placmt->maxCardServiceMove)) > 0) {
            /* cas où on on augmente le maxCardServiceMove */
            totalGain          -= delta * (long)instance.weight_SMC;
        } else if ( placmt->cardServiceMove[s1] == placmt->maxCardServiceMove && delta<0 ) {
            /* cas où on risque d'avoir diminué le maxCardServiceMove */
            recalculMax=true;
        }
    } else {
        if (placmt->cardServiceMove[s1] == placmt->maxCardServiceMove) {
            if (placmt->cardServiceMove[s2] == placmt->maxCardServiceMove) {
                /* les services diffèrent mais peuvent modifier tous deux le maxCardServiceMove */
                if (delta1 > 0 || delta2 > 0) {
                    /* cas où on augmente le maxCardServiceMove */
                    totalGain          -= (long)instance.weight_SMC;
                } else if (delta1 < 0 && delta2<0) {
                    /* cas où on risque d'avoir diminué le maxCardServiceMove */
                    newmax = (unsigned short)(placmt->maxCardServiceMove - 1)  ;
                    recalculMax=true;
                }
            } else {
                /* seul s1 peut modifier le maxCardServiceMove */
                if (delta1 > 0) {
                    /* cas où on augmente le maxCardServiceMove */
                    totalGain          -= (long)instance.weight_SMC;
                } else if (delta1 < 0) {
                    /* cas où on risque d'avoir diminué le maxCardServiceMove */
                    newmax = (unsigned short)max(placmt->maxCardServiceMove - 1, (signed)placmt->cardServiceMove[s2]+delta2) ;
                    recalculMax=true;
                }
            }
        } else if (placmt->cardServiceMove[s2] == placmt->maxCardServiceMove) {
            /* seul s2 peut modifier le maxCardServiceMove */
            if (delta2 > 0) {
                /* cas où on augmente le maxCardServiceMove */
                totalGain          -= (long)instance.weight_SMC;
            } else if (delta2 < 0) {
                /* cas où on risque d'avoir diminué le maxCardServiceMove */
                newmax = (unsigned short)max(placmt->maxCardServiceMove - 1,  ((signed)placmt->cardServiceMove[s1]+delta1)) ;
                recalculMax=true;
            }
        }
    }
    
    if (recalculMax) {
        for (i=0; i< instance.nb_services; i++) {
            if (i!=s1 && i!=s2 && newmax < placmt->cardServiceMove[i])
                newmax = placmt->cardServiceMove[i];
        }
        totalGain +=  (((long)placmt->maxCardServiceMove - newmax) * instance.weight_SMC);
    }
	//DEBUG_PRINT("SMC=%ld\n",totalGain);
	/* else pas de changement */
	return (totalGain);
}

/**
 * Calcul du cout d'une reaffectation.
 * Prend en entree le nb de machines, un pointeur vers la solution
 * initiale et un pointeur vers la solution calculee
 * et retourne le cout.
 */

int64_t cout( const placement_t *initial, const placement_t* resultat ){
    unsigned int i, j, r, r1,r2, b, m;
    unsigned target;
    int64 load_cost, load_cost_r, balance_cost, balance_cost_b, proc_move_cost, service_move_cost, machine_move_cost;
    long tempo ;

    /* Load cost: "The load cost is defined per resource
       and corresponds to the used capacity above the safety capacity;" */

    load_cost = 0;
    for( r = 0 ; r < instance.nb_ressources ; r++ ){
	load_cost_r=0;
    	for( m = 0 ; m < instance.nb_machines ; m++ ) {
		tempo = (long)resultat->U[m][r] - (long)instance.machines[m].SC[r];/* la soustraction d'unsigned donne des resultats curieux*/
		if (tempo>0) load_cost_r += (unsigned) tempo;
        }
	load_cost += instance.ressources[r].weight_loadCost * load_cost_r;
    }


    /* Balance cost */

    balance_cost = 0;


    for( b = 0 ; b < instance.nb_objectifs ; b++ ) {
	r1 = instance.objectifs[b].r1;
	r2 = instance.objectifs[b].r2;
	target = instance.objectifs[b].target;
	balance_cost_b = 0;
        for( m = 0 ; m < instance.nb_machines ; m++ ) {
		machine_t* machine = &(instance.machines[m]);
        tempo = (long) target * ((long)machine->C[r1] - (long)resultat->U[m][r1]) - ((long)machine->C[r2] -(long)resultat->U[m][r2]);
		if (tempo>0) balance_cost_b += (unsigned) tempo;
        }
	balance_cost += instance.objectifs[b].weight_balanceCost * balance_cost_b;
    }

    /* Process move cost */

    proc_move_cost = 0;

    for( i = 0 ; i < instance.nb_procs ; i++ ) {
        unsigned int depart, arrivee;
        depart = initial->assignment[i];
        arrivee = resultat->assignment[i];
        if( depart != arrivee )  {
            proc_move_cost += instance.procs[i].PMC;
        }
    }
    proc_move_cost = proc_move_cost * instance.weight_PMC;

    /* Service move cost : "a service move cost is defined as the maximum number of moved processes over services." */

    service_move_cost = 0;

    for( i = 0 ; i < instance.nb_services ; i++ ) {
        unsigned int depart, arrivee;
        unsigned int moved = 0;
        for( j = 0 ; j < instance.nb_procs ; j++ ) {
            if( UNLIKELY( i == instance.procs[j].service ) ) {
                depart = initial->assignment[j];
                arrivee = resultat->assignment[j];
                if( LIKELY( depart != arrivee ) ) {
                    moved++;
                }
            }
        }
        service_move_cost = MAX( service_move_cost, moved );
    }
    service_move_cost *= instance.weight_SMC;

    /* Machine move cost */

    machine_move_cost = 0;

    for( i = 0 ; i < instance.nb_procs ; i++ ) {
        unsigned int depart, arrivee;
        depart = initial->assignment[i];
        arrivee = resultat->assignment[i];
        if( LIKELY( depart != arrivee ) ) {
            machine_move_cost += instance.machines[ depart ].MMC[arrivee ];
        }

    }
    machine_move_cost *= instance.weight_MMC;

    DEBUG_PRINT( "Load cost: %lld, balance cost: %lld, process move cost: %lld, service move cost: %lld, machine move cost: %lld\n", load_cost, balance_cost, proc_move_cost, service_move_cost, machine_move_cost );

    /* Cout total */

    return(load_cost + balance_cost + proc_move_cost +  service_move_cost +  machine_move_cost);
}


