/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/

#include <string.h>
#include <stdint.h>
#include <values.h>

#include "fmr.h"
#include "fmr_mem.h"
#include "fmr_tools.h"
#include "fmr_constraints.h"
#include "fmr_psmvt.h"
#include "fmr_cost.h"
#include "fmr_glouton.h"
#include "fmr_threads.h"
#include "fmr_vns.h"
#include "fmr_exploSwap.h"
#include "fmr_exploShift.h"
#include "fmr_exploEjection.h"

typedef struct {
    double ratio;
    unsigned short id;   /* id du ps */
} ratio_t;

// to be used by qsort
static int
compare_ratio_t(const void *a, const void *b)
{
    const ratio_t *ra = (const ratio_t *)a;
    const ratio_t *rb = (const ratio_t *)b;
    if (ra->ratio < rb->ratio)
        return(-1);
    else if (ra->ratio == rb->ratio)
        return(0);
    else
        return(1);
}


double
normeL1(double avant, unsigned short int p, unsigned short int m, unsigned short int r, unsigned c) {
    if (c)
    {
//         DEBUG_PRINT("normeL1 avant=%lf inst.procs.R=%d c=%d return=%lf\n", avant, instance.procs[p].R[r], c, avant + (double)instance.procs[p].R[r] / c);
        return (avant + (double)instance.procs[p].R[r] / c);
    }
    else
        return (MAXDOUBLE); // une sorte de infini
}

double
normeLinf(double avant, unsigned short int p, unsigned short int m, unsigned short int r, unsigned c) {
    if (c) 
        return ( MAX(avant, (double)instance.procs[p].R[r] / c));
    else
        return (MAXDOUBLE); // une sorte de infini
}

bool
placement_glouton_partiel (const placement_t* initial, placement_t* solution, double (norme)(double,unsigned short int,unsigned short int,unsigned short int,unsigned int), liste_t *ps_aplacer, liste_t *machs) {
    ratio_t *tabRatios = (ratio_t *) malloc( instance.nb_procs * sizeof(ratio_t) );
    unsigned short i, p,pred, m,m0,r,best_mach, compteur;
    double maxcout, cout;
    unsigned cap;
    
    /* tente l'affectation selon l'ordre décroissant des ratios, si un ps
     * ne peut être placé, on le remonte par un échange avec le ps de ratio sup.
     * et ce tant que l'on a pas tout essayé.*/
    
    
    /* on pose tout de suite ceux qui n'ont pas de choix autre que leur m0 */
    p=liste_head(ps_aplacer);
    while (p!= LISTE_VIDE_US) {
        m=liste_head(machs);
        compteur=0;
        while (m!= LISTE_VIDE_US) {
            if(check_add_proc_glouton2 (initial, p, m, solution)) {
                compteur++;
            }
            m = liste_next(m, machs);
        }
        if (compteur==1) { /* forcément m0 */
            ajouter_proc_glouton(initial, p, initial->assignment[p], solution);
            DEBUG_PRINT("placement forcé de %hu sur %hu\n",p, initial->assignment[p]);
            pred=p;
            p = liste_next(p, ps_aplacer);	
            liste_out(pred, ps_aplacer);
        } else {
            p = liste_next(p, ps_aplacer);	
        }
    }
    
    /* On fixe sur leur m0 les gros processus suivant au moins 1 transient */
    
    p=liste_head(ps_aplacer);
    while (p!= LISTE_VIDE_US) {
        m0= initial->assignment[p];
        pred=LISTE_VIDE_US;
        for(r=0;r<instance.nb_ressources;r++) {
            if (instance.ressources[r].transcient) {
                if (instance.procs[p].R[r] > 0.3*(instance.machines[m0].C[r]-initial->U[m0][r])) {
                    ajouter_proc_glouton(initial, p, m0, solution);
                    DEBUG_PRINT("placement conseillé de %hu sur %hu\n",p, initial->assignment[p]);
                    pred=p;
                    break;
                }
            }
        }
        p = liste_next(p, ps_aplacer);
        if (pred!=LISTE_VIDE_US) {
            liste_out(pred, ps_aplacer);
        }
    }
    
    #ifdef DEBUG
    p=liste_head(ps_aplacer);
    compteur=0;
    while (p!= LISTE_VIDE_US) {
        compteur++;
        DEBUG_PRINT("%d,",p);
        p = liste_next(p, ps_aplacer);
    }
    DEBUG_PRINT("reste %d ps\n avant placement par ratio\n", compteur);
    #endif
    
    /* Calcul des ratios */
    compteur=0;
    p=liste_head(ps_aplacer);
    while (p!= LISTE_VIDE_US) {
        tabRatios[compteur].ratio=0.0;
        tabRatios[compteur].id=p;
        m0 = initial->assignment[p];
        m=liste_head(machs);
        while (m!= LISTE_VIDE_US) {
            for(r=0;r<instance.nb_ressources;r++) {
                cap = instance.machines[m].C[r];
                if(instance.ressources[r].transcient) {
                    if(m != m0) {
                        cap = instance.machines[m].C[r] - initial->U[m][r];
                    } else {
                        cap = UINT32_MAX;
                    }
                }
                tabRatios[compteur].ratio = norme(tabRatios[compteur].ratio,p,m,r,cap);
            }
            m = liste_next(m, machs);
        }
        p = liste_next(p, ps_aplacer);
        compteur++;
    }
    
    /* qsort trie par ordre croissant */
    qsort(tabRatios, compteur, sizeof(ratio_t), compare_ratio_t);
    
    #ifdef DEBUG
    for(i=0; i<compteur; i++) 
        DEBUG_PRINT("%d:%lf,",tabRatios[i].id, tabRatios[i].ratio);
    DEBUG_PRINT("\n");
    #endif
    
    
    best_mach=LISTE_VIDE_US;
    for (i=(unsigned short)(compteur -1); i!=LISTE_VIDE_US; i--) {
        p=tabRatios[i].id;
        maxcout= LONG_MIN;
        best_mach=LISTE_VIDE_US;
        m=liste_head(machs);
        while (m!= LISTE_VIDE_US) {
            if(check_add_proc_glouton (initial, p, m, solution)) {
                cout=(double) calcule_gain_ajout(initial, solution, p, m);
                if(cout>maxcout) {
                    maxcout=cout;
                    best_mach=m;
                }
            }
            m = liste_next(m, machs);
        }
        if (best_mach == LISTE_VIDE_US) {
            DEBUG_PRINT("On n'arrive pas à placer le ps %hu (m0=%hu)!\n", p, initial->assignment[p]);
            break; /* pas la peine d'aller plus loin */
        } else {
            ajouter_proc_glouton (initial, p, best_mach, solution);
            compteur--;
            DEBUG_PRINT("1) compteur:%d proc:%hu de service:%hu sur machine:%hu au lieu de %hu\n", compteur, p, instance.procs[p].service, best_mach, initial->assignment[p]);
        }
    }
    DEBUG_PRINT("nbre de ps non placés:%hu\n", compteur);
    free(tabRatios);
    if (best_mach== LISTE_VIDE_US && compteur>0)
        return(false);
    return(true);
}

long
placement_glouton_partiel2 (const placement_t* initial, placement_t* solution, double (norme)(double,unsigned short int,unsigned short int,unsigned short int,unsigned int), liste_t *ps_aplacer) {
    ratio_t *tabRatios = (ratio_t *) malloc( instance.nb_procs * sizeof(ratio_t) );
    unsigned short i, p,pred, m,m0,r,best_mach, compteur;
    long maxgain=0, gain, gainTotal=0;
    unsigned cap;

    /* tente l'affectation selon l'ordre décroissant des ratios, si un ps
     * ne peut être placé, on le remonte par un échange avec le ps de ratio sup.
     * et ce tant que l'on a pas tout essayé.*/
    
    
    /* on pose tout de suite ceux qui n'ont pas de choix autre que leur m0 */
    /*
    p=liste_head(ps_aplacer);
    while (p!= LISTE_VIDE_US) {
        compteur=0;
	for (m=0; m< instance.nb_machines; m++) {
            if(check_add_proc_glouton2 (initial, p, m, solution)) {
                compteur++;
            }
        }
        if (compteur==1) { * forcément m0 *
            ajouter_proc_glouton(initial, p, initial->assignment[p], solution);
            DEBUG_PRINT("placement forcé de %hu sur %hu\n",p, initial->assignment[p]);
            pred=p;
            p = liste_next(p, ps_aplacer);	
            liste_out(pred, ps_aplacer);
        } else {
            p = liste_next(p, ps_aplacer);	
        }
    }
*/ 
    /* On fixe sur leur m0 les gros processus suivant au moins 1 transient */
 /*   
    p=liste_head(ps_aplacer);
    while (p!= LISTE_VIDE_US) {
        m0= initial->assignment[p];
        pred=LISTE_VIDE_US;
        for(r=0;r<instance.nb_ressources;r++) {
            if (instance.ressources[r].transcient) {
                if (instance.procs[p].R[r] > 0.3*(instance.machines[m0].C[r]-initial->U[m0][r])) {
                    ajouter_proc_glouton(initial, p, m0, solution);
                    DEBUG_PRINT("placement conseillé de %hu sur %hu\n",p, initial->assignment[p]);
                    pred=p;
                    break;
                }
            }
        }
        p = liste_next(p, ps_aplacer);
        if (pred!=LISTE_VIDE_US) {
            liste_out(pred, ps_aplacer);
        }
    }
  */  
    #ifdef DEBUG
    p=liste_head(ps_aplacer);
    compteur=0;
    while (p!= LISTE_VIDE_US) {
        compteur++;
        DEBUG_PRINT("%d,",p);
        p = liste_next(p, ps_aplacer);
    }
    DEBUG_PRINT("reste %d ps\n avant placement par ratio\n", compteur);
    #endif
    
    /* Calcul des ratios */
    compteur=0;
    p=liste_head(ps_aplacer);
    while (p!= LISTE_VIDE_US) {
        tabRatios[compteur].ratio=0.0;
        tabRatios[compteur].id=p;
        m0 = initial->assignment[p];
        for (m=0; m<instance.nb_machines;m++) {
            for(r=0;r<instance.nb_ressources;r++) {
                cap = instance.machines[m].C[r];
                if(instance.ressources[r].transcient) {
                    if(m != m0) {
                        cap = instance.machines[m].C[r] - initial->U[m][r];
                    } else {
                        cap = UINT32_MAX;
                    }
                }
                tabRatios[compteur].ratio = norme(tabRatios[compteur].ratio,p,m,r,cap);
            }
        }
        p = liste_next(p, ps_aplacer);
        compteur++;
    }
    
    /* qsort trie par ordre croissant */
    qsort(tabRatios, compteur, sizeof(ratio_t), compare_ratio_t);
    
    #ifdef DEBUG
    for(i=0; i<compteur; i++) 
        DEBUG_PRINT("%d:%lf,",tabRatios[i].id, tabRatios[i].ratio);
    DEBUG_PRINT("\n");
    #endif
    
    
    best_mach=LISTE_VIDE_US;
    for (i=(unsigned short)(compteur -1); i!=LISTE_VIDE_US; i--) {
        p=tabRatios[i].id;
        maxgain= LONG_MIN;
        best_mach=LISTE_VIDE_US;
	for (m=0; m<instance.nb_machines;m++) {
            if(check_add_proc_glouton (initial, p, m, solution)) {
                gain = calcule_gain_ajout(initial, solution, p, m);
                if(gain >maxgain) {
                    maxgain=gain;
                    best_mach=m;
                }
            }
        }
        if (best_mach == LISTE_VIDE_US) {
            DEBUG_PRINT("On n'arrive pas à placer le ps %hu (m0=%hu)!\n", p, initial->assignment[p]);
            break; /* pas la peine d'aller plus loin */
        } else {
            ajouter_proc_glouton (initial, p, best_mach, solution);
            compteur--;
		DEBUG_PRINT("gain=%ld\n",gain);
            //DEBUG_PRINT("Compteur:%d proc:%hu de service:%hu sur machine:%hu au lieu de %hu\n", compteur, p, instance.procs[p].service, best_mach, initial->assignment[p]);
        }
    }
    DEBUG_PRINT("nbre de ps non placés:%hu, gainTotal=%ld\n", compteur,gainTotal);
    free(tabRatios);
    if (best_mach== LISTE_VIDE_US && compteur>0)
        return(LONG_MIN);
    return(0);
}

long explo_glouton( const placement_t const * initial, placement_t* resultat )
{
    placement_t glouton;
    glouton = alloc_assignement( instance.nb_procs );
    init_vide(&glouton);
    if ( placement_glouton(initial, &glouton, normeL1) && check_all_constraints(initial,&glouton))
    {
	atomic_set_meilleur_placement(&glouton, 'G');
        copier_assignement (&glouton, resultat);
        resultat->cout = cout(initial, resultat);
        DEBUG_PRINT("Glouton réussi : coût de la solution : %lld\n", resultat->cout);
    } else {
        DEBUG_PRINT("Glouton raté\n");
    }
    free_placement(&glouton);
    return (long) (initial->cout - resultat->cout);
}

bool
placement_glouton(const placement_t* initial, placement_t* solution, double (norme)(double,unsigned short int,unsigned short int,unsigned short int,unsigned int)){
    unsigned short int m, r;
    bool saturee;
    liste_t ps_aplacer;
    unsigned short i;
    liste_t machines_dispo;
    
    liste_init(&ps_aplacer, instance.nb_procs);
    liste_remplie(&ps_aplacer); /* sauf preuve du contraire tous sont à placer */ 
    liste_init(&machines_dispo, instance.nb_machines);
    liste_remplie(&machines_dispo);
    
    for(m=0;m<instance.nb_machines;m++)
    {
        saturee=false;
        for(r=0;r<instance.nb_ressources;r++) {
            if (instance.ressources[r].transcient && instance.machines[m].C[r] == initial->U[m][r]) {
                saturee=true;
                break;
            }
        }
        if (saturee) {
            DEBUG_PRINT("--> %hu saturee par transcient\n", m);
	    i= listeCompacte_tete( &(initial->liste_procs), m );
            while (i!= LISTE_VIDE_US) {
                ajouter_proc_glouton(initial, i, m, solution);
                liste_out(i, &ps_aplacer);
		i= listeCompacte_next(&initial->liste_procs, m, i);
            }
            liste_out(m, &machines_dispo);
        }
    }
    return placement_glouton_partiel(initial, solution, normeL1, &ps_aplacer, &machines_dispo);
}

long explo_rap_glouton( const placement_t const * initial, placement_t* resultat )
{
    placement_t rap_glouton;
    rap_glouton = alloc_assignement( instance.nb_procs );
    copier_assignement (resultat, &rap_glouton);

    if ( remove_add_procs_glouton(initial, &rap_glouton, normeL1) && check_all_constraints(initial, &rap_glouton))
    {
	atomic_set_meilleur_placement(&rap_glouton, 'D');
        copier_assignement (&rap_glouton, resultat);
        resultat->cout = cout(initial, resultat);
        DEBUG_PRINT("Rap Glouton réussi : coût de la solution : %lld\n", resultat->cout);
    } else {
        DEBUG_PRINT("Rap Glouton raté\n");
    }
    free_placement(&rap_glouton);
    return (long) (initial->cout - resultat->cout);
}

liste_t 
reduction_glouton(const placement_t* initial, placement_t* solution, unsigned short absent_ou_indice[], double (norme)(double,unsigned short int,unsigned short int,unsigned short int,unsigned int)){
    unsigned short i;
    unsigned short int m, p, compteur, der_present= 0;
    liste_t ps_aplacer, *l;
	/*  absent_ou_indice[i]=NOPROCESS si présent sinon indice du prochain présent (i.e. non oté), 0 si pas de prochain présent */

    liste_init(&ps_aplacer, instance.nb_procs);

    /*On retire des processus*/
    compteur=0;
    for (p =  (unsigned short)(instance.nb_procs-1); p!= NOPROCESS ; p--) {
	m = solution->assignment[p];
	if (check_remove_proc (p, m, solution)) {	/* si une fois p supprimé la sol. est toujours valide */
	  liste_in( p, &ps_aplacer);
	  oter_proc_glouton (initial, p, m, solution);
	  compteur++;
	  absent_ou_indice[p]=der_present;
	} else {
		der_present=p;
		absent_ou_indice[p]=NOPROCESS;
	}
    }  

    DEBUG_STATS_PRINT("reduction_Glouton nb procs enleves : %d (coeur de %d)\n", compteur, instance.nb_procs-compteur);

    return ps_aplacer;
}

long
explo_GloutonReduction (const placement_t* initial, placement_t* solution, long gainMin) {
	unsigned short *absent_ou_indice= (unsigned short *)malloc(sizeof(unsigned short)*instance.nb_procs);
	liste_t ps_aplacer;
// 	unsigned short longChaine = MAXLONGCHAINEEJECTION;	/* choisi par experimentation */
	long gain=0, gainPlacement;
	unsigned short m, p;
    	bool saturee;
	bool *machine_saturee = (bool *) malloc(sizeof(unsigned short)*instance.nb_machines);
	placement_t sauve_sol;

#ifdef DEBUGSTATS
        printf("avant copie placement temps=%lf\n",tempsEcoule2());
#endif
        sauve_sol= alloc_assignement(instance.nb_procs);
	copier_assignement(solution, &sauve_sol);

#ifdef DEBUGSTATS
        printf("Après copie placement temps=%f\n",tempsEcoule2());
#endif

	ps_aplacer = reduction_glouton(initial,solution, absent_ou_indice, normeL1);

#ifdef DEBUGSTATS
	printf("Après reduction_glouton temps=%f\n",tempsEcoule2());
#endif

// 	gain+=explo_voisinageSwap_restreint(initial, solution, absent_ou_indice, gainMin);
// 	printf("gain Swap restreint=%ld\n",gain);
// 	gain+=explo_voisinageBestSwap_restreint(initial, solution, absent_ou_indice, gainMin);
// 	printf("gain BestSwap restreint=%ld\n",gain);
// 	gain+=explo_voisinageVeryBestSwap_restreint(initial, solution, absent_ou_indice, gainMin);
// 	printf("gain VeryBestSwap restreint=%ld\n",gain);
	gain+=explo_voisinageShift_restreint(initial, solution, absent_ou_indice, gainMin);
	DEBUG_PRINT("gain Shift restreint=%ld\n",gain);
// 	gain+=explo_voisinageBestShift_restreint(initial, solution, absent_ou_indice, gainMin);
// 	printf("gain BestShift restreint=%ld\n",gain);
// 	gain+=explo_voisinageVeryBestShift_restreint(initial, solution, absent_ou_indice, gainMin);
// 	printf("gain VeryBestShift restreint=%ld\n",gain);
// 	gain+=explo_voisinageEjectionChain_restreint(initial, solution, longChaine, absent_ou_indice, gainMin); //ne fonctionne pas car utilise check_remove_proc
// 	printf("gain EjectionChain restreint=%ld\n",gain);

#ifdef DEBUGSTATS
	printf("Après voisinageSwap_restreint temps=%f\n",tempsEcoule2());
	printf("gain Swap_restreint=%ld ",gain);
#endif

    	if (placement_glouton_partiel2(initial, solution, normeL1, &ps_aplacer)>LONG_MIN) {
		solution->cout=cout(initial,solution);
		DEBUG_PRINT("gain sur explo_GloutonReduction:%llu\n", initial->cout - solution->cout);
		return(initial->cout - solution->cout);
	} else {
		copier_assignement(&sauve_sol,solution);
		DEBUG_STATS_PRINT("glouton raté\n");
		return(LONG_MIN);
	}
	free(absent_ou_indice);
	free_placement(&sauve_sol);
}

bool
remove_add_procs_glouton(const placement_t* initial, placement_t* solution, double (norme)(double,unsigned short int,unsigned short int,unsigned short int,unsigned int)){
    unsigned short i;
    unsigned short int m, r, p, compteur;
    bool saturee;
    int64 gain;
    liste_t ps_aplacer, *l;
    liste_t machines_dispo;
    unsigned long oteproc[instance.nb_procs];
    unsigned long machC[instance.nb_machines];
    unsigned int coef=instance.nb_procs/80; /* coef a regler pour accepter de retirer plus ou moins de processus (0 = tous les procs possibles)*/
/*    placement_t *initial_relax;
    unsigned int seed=16;*/

    liste_init(&ps_aplacer, instance.nb_procs);
    /*calculs initiaux sur les machines et processus pour choisir quels processus enlever*/
    for( m = 0 ; m < instance.nb_machines ; m++ ) {
        machC[m]=0;
        for( r = 0 ; r < instance.nb_ressources ; r++ ) {
            machC[m]+=instance.machines[m].C[r];
        }
    }
    for( p = 0 ; p < instance.nb_procs ; p++ ) {
        oteproc[p]=0;
        for( r = 0 ; r < instance.nb_ressources ; r++ ) {
            oteproc[p]+=instance.procs[p].R[r];
        }
    }
    /*On retire des processus*/
    compteur=0;
//     for (p = 0; p < instance.nb_procs && (((float)compteur/(float)instance.nb_procs)<0.8); p++) {
    for (p = 0; p < instance.nb_procs; p++) {
	m = solution->assignment[p];
	if (check_remove_proc (p, m, solution) && (coef*instance.nb_ressources*oteproc[p]<machC[m])) {	/* si une fois p supprimé la sol. est toujours valide */
	  liste_in( p, &ps_aplacer);
	  oter_proc_glouton (initial, p, m, solution);
	  compteur++;
	}
    }  
    DEBUG_PRINT("Remove add procs Glouton nb procs enleves : %d\n", compteur);
//     gain = calcul_assignement_vns (initial, initial_relax, solution, &seed);    
    liste_init(&machines_dispo, instance.nb_machines);
    liste_remplie(&machines_dispo);
    
    for(m=0;m<instance.nb_machines;m++)
    {
        saturee=false;
        for(r=0;r<instance.nb_ressources;r++) {
            if (instance.ressources[r].transcient && instance.machines[m].C[r] == initial->U[m][r]) {
                saturee=true;
                break;
            }
        }
        if (saturee) {
            DEBUG_PRINT("--> %u saturee par transcient\n", m);
            //l= & initial->liste_procs[m];
            //i= liste_head(l);
		i=listeCompacte_tete(& initial->liste_procs, m);
            while (i!= LISTE_VIDE_US) {
	      	if(liste_appartient(i, &ps_aplacer)) {
                	ajouter_proc_glouton(initial, i, m, solution);
                	liste_out(i, &ps_aplacer);
	      	}
		i=listeCompacte_next(& initial->liste_procs, m,i);
            }
            liste_out(m, &machines_dispo);
        }
    }
    return placement_glouton_partiel(initial, solution, normeL1, &ps_aplacer, &machines_dispo);
}
