/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/

#include "fmr_mem.h"
#include "fmr.h"
#include "fmr_tools.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h> /* pour gettimeofday() */
#include <unistd.h>   /* pour getpid() */
#include <stdarg.h>   /* pour vararg */
#include <stdint.h>


/* marqueur d'init de la structure: -2 sur le premier assign */
bool
placement_est_initialise (placement_t *pl) {
	return(pl !=NULL && pl->assignment[0]!= NOINIT);
}

/* Copier un placement dans un autre */
void copier_assignement( const placement_t const *initial, placement_t *arrivee) {
    unsigned short n,m,s;

    if (placement_est_initialise(arrivee)) {
	listeCompacte_cpy( & arrivee->liste_procs, & initial->liste_procs, instance.nb_machines, instance.nb_procs);
	listeCompacte2D_cpy(& (arrivee->liste_de_processus_d_un_service_dans_un_voisinage), 
			    	&(initial->liste_de_processus_d_un_service_dans_un_voisinage), 
				instance.nb_neighborhood, instance.nb_services,
				instance.nb_procs);

#ifdef TRACE_ALGOS
	arrivee->liste_algos = (char *) realloc( arrivee->liste_algos, sizeof(char)*(initial->nb_algos+1) );
	memcpy( arrivee->liste_algos, initial->liste_algos, initial->nb_algos+1 );
	arrivee->nb_algos  = initial->nb_algos ;
#endif
    } else {
	listeCompacte_alloc( & arrivee->liste_procs,  instance.nb_machines, instance.nb_procs);
	listeCompacte_cpy( & arrivee->liste_procs, & initial->liste_procs, instance.nb_machines, instance.nb_procs);

	listeCompacte2D_alloc(& (arrivee->liste_de_processus_d_un_service_dans_un_voisinage),  instance.nb_neighborhood, instance.nb_services, instance.nb_procs);
	listeCompacte2D_cpy(& (arrivee->liste_de_processus_d_un_service_dans_un_voisinage), 
				&(initial->liste_de_processus_d_un_service_dans_un_voisinage), instance.nb_neighborhood, instance.nb_services,
                               	instance.nb_procs);

#ifdef TRACE_ALGOS
	arrivee->liste_algos = (char *) malloc( initial->nb_algos+1 );
	memcpy( arrivee->liste_algos, initial->liste_algos, initial->nb_algos+1 );
	arrivee->nb_algos  = initial->nb_algos ;
#endif
    }

    memcpy( arrivee->assignment, initial->assignment, instance.nb_procs * sizeof( unsigned short )  );
    memcpy( arrivee->cardServiceMove, initial->cardServiceMove, instance.nb_services* sizeof( unsigned short ) );

    array2Dcpy (arrivee->U, initial->U, sizeof(unsigned), instance.nb_machines, instance.nb_ressources);
    array2Dcpy (arrivee->residu, initial->residu, sizeof(unsigned), instance.nb_machines, instance.nb_ressources);
    array2Dcpy (arrivee->service_par_machine, initial->service_par_machine, sizeof(unsigned short), instance.nb_machines, instance.nb_services);
    array2Dcpy (arrivee->spread, initial->spread, sizeof(unsigned short), instance.nb_services, instance.nb_locations);

    memcpy ( arrivee->nb_procs_par_mach, initial->nb_procs_par_mach, instance.nb_machines * sizeof(unsigned short));
    arrivee->maxCardServiceMove = initial->maxCardServiceMove;

    arrivee->cout = initial->cout;
}


/* Allouer la structure de donnees contenant un placement */
placement_t alloc_assignement( unsigned short nb_procs ) {
    placement_t pl;

    pl.assignment = (unsigned short*) malloc( nb_procs * sizeof(unsigned short) );
    pl.assignment[0] = NOINIT; /* marqueur pour verifier l'initialisation de cette structure placement */
    pl.cardServiceMove = (unsigned short*) malloc( instance.nb_services * sizeof( unsigned short ) );

    pl.U = (unsigned **) array2Dcalloc(sizeof(unsigned), instance.nb_machines, instance.nb_ressources);
    pl.residu = (unsigned **) array2Dcalloc(sizeof(unsigned), instance.nb_machines, instance.nb_ressources);
    pl.service_par_machine = (unsigned short **) array2Dmalloc(sizeof(unsigned short), instance.nb_machines, instance.nb_services);
    
    pl.spread = (unsigned short **) array2Dmalloc(sizeof(unsigned short), instance.nb_services, instance.nb_locations);

    pl.nb_procs_par_mach = (unsigned short *) malloc (instance.nb_machines * sizeof(unsigned short));
	
    listeCompacte_alloc( & pl.liste_procs, instance.nb_machines, instance.nb_procs);
    listeCompacte2D_alloc(& (pl.liste_de_processus_d_un_service_dans_un_voisinage),  instance.nb_neighborhood, instance.nb_services, instance.nb_procs);
    pl.maxCardServiceMove=0; // pour éviter un stupide warning
    pl.cout = INT64_MAX ;
#ifdef TRACE_ALGOS
	pl.liste_algos = (char *) malloc( 1 * sizeof(char) );
	pl.liste_algos[0] = 0 ;
	pl.nb_algos  = 0 ;
#endif

    return pl;
}

/* Liberation d'un placement */
void free_placement( placement_t* pl ) {
	free( pl->U);	/* avantage de array2Dcalloc : pas de boucle */
	free( pl->residu);
	free( pl->service_par_machine);
	free( pl->spread);
	free( pl->nb_procs_par_mach);
	free( pl->cardServiceMove);
	listeCompacte_free(&pl->liste_procs);
	//free(pl->liste_de_processus_d_un_service_dans_un_voisinage);
	listeCompacte2D_free(& pl->liste_de_processus_d_un_service_dans_un_voisinage);
        free( pl->assignment );
#ifdef TRACE_ALGOS
	free(pl->liste_algos);
#endif
}


/**
 * Creer une solution initiale vide.
 * Retourne FMR_OK si tout s'est bien passe, une erreur si il
 * y a eu un probleme.
 */
void
init_vide(placement_t* solution){
    unsigned int n,m,s,l;

    memset (solution->assignment,(signed)NOMACHINE, instance.nb_procs *sizeof(unsigned short));
    memset(solution->cardServiceMove, 0, instance.nb_services *sizeof(unsigned short));
    memset(solution->nb_procs_par_mach, 0, instance.nb_machines *sizeof(unsigned short));

    /* U et residu deja initialisés à 0 par calloc */

    /* Si NOSERVICE=-1u alors = 0xFFFF sur un unsigned short, donc on  peut utiliser memset 
     * grâce aux array2D */
    memset (solution->service_par_machine[0], 0xFF, (size_t)instance.nb_machines*instance.nb_services*sizeof(unsigned short));

    listeCompacte_init( & solution->liste_procs, instance.nb_machines, instance.nb_procs);

    memset (solution ->spread[0], 0, (size_t)instance.nb_services*instance.nb_locations*sizeof(unsigned short));

    listeCompacte2D_init(& solution->liste_de_processus_d_un_service_dans_un_voisinage, instance.nb_neighborhood, 
			instance.nb_services, instance.nb_procs);
}
