/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/

#include "fmr.h"
#include "fmr_tools.h"
#include "fmr_constraints.h"
#include "fmr_psmvt.h"


/* enlever le processus p du placement, suppose que les vérifs de contraintes
   on déjà été faites, voir check_remove_proc */
void
oter_proc (const placement_t const *initial, unsigned short p, unsigned short m,
	   placement_t * placmt)
{
    unsigned short i, r;
    unsigned short s = instance.procs[p].service;
    unsigned short l = instance.machines[m].location;
    unsigned short m0 = initial->assignment[p];
    unsigned short n = instance.machines[m].neighborhood;
    unsigned short newmax;

     placmt->assignment[p]=NOMACHINE;
    /* mise à jour du taux de charge, U et residu ne peuvent etre fusionnés */
    for (r = 0; r < instance.nb_ressources; r++) {
	if (m== m0 && instance.ressources[r].transcient )
		placmt->residu[m][r] += instance.procs[p].R[r];
	placmt->U[m][r] -= instance.procs[p].R[r];
    }

    listeCompacte_out(&placmt->liste_procs, m, p);
    /*if (liste_out (p, &(placmt->liste_procs[m])) == LISTE_VIDE) {
	DEBUG_PRINT ("erreur p non trouvé dans ote_proc\n");
	exit (FMR_NOT_FOUND);
    }*/

    placmt->nb_procs_par_mach[m]--;
    placmt->spread[s][l]--;
    placmt->service_par_machine[m][s] = NOPROCESS;

    if (m != m0) {
	placmt->cardServiceMove[s]--;
	if (placmt->maxCardServiceMove == (unsigned short)(placmt->cardServiceMove[s] + 1)) {	/* donc il faut recalculmax */
	    //recalcul_maxCardServiceMove(initial, placmt, s, -1);
	    newmax = 0;
	    for (i = 0; i < instance.nb_services; i++) {
		if (placmt->cardServiceMove[i] > newmax)
		    newmax = placmt->cardServiceMove[i];
	    }
	    placmt->maxCardServiceMove = newmax;
	}
    }

    /* mise à jour voisinage_par_service */
    //if (liste_out (p, &(placmt->liste_de_processus_d_un_service_dans_un_voisinage[n][s])) == LISTE_VIDE) {
    listeCompacte2D_out(&(placmt->liste_de_processus_d_un_service_dans_un_voisinage), n,s,p);
	//fprintf (stderr, "ps %d pas dans le voisinage n=%d, s=%d!\n", p, n, s), exit (1);
    //}
}


/* ajouter le processus p à la machine m, suppose que les vérifs de contraintes
   ont déjà été faites */
void
ajouter_proc (const placement_t const *initial, unsigned short p, unsigned short m, placement_t * placmt)
{
    unsigned short r;
    unsigned short s = instance.procs[p].service;
    unsigned short l = instance.machines[m].location;
    unsigned short m0 = initial->assignment[p];
    unsigned short n = instance.machines[m].neighborhood;

    placmt->assignment[p]= m;
    /* mise à jour du taux de charge, U et residu ne peuvent etre fusionnés */
    for (r = 0; r < instance.nb_ressources; r++) {
	if (instance.ressources[r].transcient) {
	    if (m0 == m)	/*  depart de la machine initiale */
		placmt->residu[m0][r] -= instance.procs[p].R[r];
	}
	placmt->U[m][r] += instance.procs[p].R[r];
    }

    listeCompacte_in( &(placmt->liste_procs), m, p );

    placmt->nb_procs_par_mach[m]++;
    placmt->spread[s][l]++;
#ifdef DEBUG
    if (placmt->service_par_machine[m][s] != NOPROCESS) {
	DEBUG_PRINT
	    ("pb il y a deja un ps (%hu) de service %hu sur machine %hu dans ajouter_proc_bis(p=%hu)\n",
	     placmt->service_par_machine[m][s], s, m, p);
	exit (1);
    }
#endif
    placmt->service_par_machine[m][s] = p;

    if (m != m0) {
	if (placmt->maxCardServiceMove == placmt->cardServiceMove[s]) {
	    placmt->maxCardServiceMove++;
	}
	placmt->cardServiceMove[s]++;
    }
    listeCompacte2D_in ( &(placmt->liste_de_processus_d_un_service_dans_un_voisinage), n,s,p);
}
/* Ajouter un proc sur une machine
 * Arguments : instance, placement, rang de la machine, rang du processus
 * et un booleen qui dit si on force l'ajout ou pas si la machine est surchargee
 * Retourne FMR_OK si c'est bon et FMR_OVER si la machine est surchargeee
 * ATTENTION PAS A JOUR DE LA DERNIERE VERSION DE placement_t
 */

int ajouter_proc_old( placement_t* placmt, unsigned int m, unsigned int rk_proc, bool force ) {

    machine_t* machine;
    processus_t* process;
    unsigned int r;

    machine = &( instance.machines[ m ] );
    process = &( instance.procs[ rk_proc ] );

    /* La machine sera-t-elle surchargee si on ajoute ce proc */

    for( r = 0 ; r < instance.nb_ressources ; r++ ) {
        if( ( placmt->U[m][r] + process->R[r] > machine->C[r] )
            && ( false == force ) ){
            DEBUG_PRINT( " machine %d chargee a %d / %d, on demande %d en plus : c\'est trop\n",
		m,placmt->U[m][r], machine->C[r], process->R[r] );
            return FMR_OVERLOAD;
        }
    }
    DEBUG_PRINT( "Charge de la machine %d avant ajout: CPU %d/%d, RAM %d/%d\n",
		    m, placmt->U[m][0], machine->C[0], placmt->U[m][1], machine->C[1] );

    /* Update de l'assignement */

    placmt->assignment[ rk_proc ] = (unsigned short)m;

    /* Update des charges des ressources des machines */

    for( r = 0 ; r < instance.nb_ressources ; r++ ) {
        placmt->U[m][r] += process->R[r];
    }

    DEBUG_PRINT( "Charge de la machine %d apres ajout: CPU %d, RAM %d\n", m, placmt->U[m][0], placmt->U[m][1] );

    /* Ajouter le processus sur la machine */

    listeCompacte_in( &placmt->liste_procs,m,(unsigned short)rk_proc);
    placmt->nb_procs_par_mach[m]++;

    return FMR_OK;
}


/* Retirer un proc d'une machine
 * Arguments : pointeur vers l'instance, rang de la machine, rang du processus
 * Retourne FMR_OK si c'est bon et FMR_NOT_FOUND si le processus
 * n'est pas sur la machine (ne devrait pas arriver)
*/

int retirer_proc( placement_t* placmt, unsigned short m, unsigned short p ) {
    machine_t* machine;
    processus_t* process;
    unsigned short r;

    machine = &( instance.machines[ m ] );
    process = &( instance.procs[ p ] );

    DEBUG_PRINT( "Charge de la machine %d avant : CPU %d, RAM %d\n", m, placmt->U[m][0], placmt->U[m][1] );

    /* Update des charges des ressources des machines */

    for( r = 0 ; r < instance.nb_ressources ; r++ ) {
        placmt->U[m][r] -= process->R[r];
    }

    DEBUG_PRINT( "Charge de la machine %d apres: CPU %d, RAM %d\n", m, placmt->U[m][0], placmt->U[m][1] );

    /* Retirer le processus de la machine */

    listeCompacte_out(& placmt->liste_procs, m,(unsigned short) p);

    if( NULL == process ) {
        return FMR_NOT_FOUND;
    }

    placmt->nb_procs_par_mach[m]--;

    return FMR_OK;

}

/* Echanger deux processus */

int swap_procs( placement_t* placmt, unsigned short m1, unsigned short p1, unsigned short m2, unsigned short p2, bool force ) {

    machine_t* machine1;
    machine_t* machine2;
    processus_t* process1;
    processus_t* process2;
    unsigned short r;

    machine1 = &( instance.machines[ m1 ] );
    machine2 = &( instance.machines[ m2 ] );
    process1 = &( instance.procs[ p1 ] );
    process2 = &( instance.procs[ p2 ] );

    /* La machine 2 sera-t-elle surchargee si on ajoute ce proc et on retire l'autre */

    for( r = 0 ; r < instance.nb_ressources ; r++ ) {
        if( ( placmt->U[m2][r] + process1->R[r] - process2->R[r] > machine2->C[r] )
             && ( false == force ) ){
           //DEBUG_PRINT( " machine chargee a %d / %d, on demande %d en plus : c\'est trop\n",
	//		   placmt->U[m2][r], machine2->C[r], process1->R[r] );
            return FMR_OVERLOAD;
        }
    }

    /* La machine 1 sera-t-elle surchargee si on ajoute ce proc et on retire l'autre */

    for( r = 0 ; r < instance.nb_ressources ; r++ ) {
        if( ( placmt->U[m1][r] + process2->R[r] - process1->R[r] > machine1->C[r] )
             && ( false == force ) ){
          // DEBUG_PRINT( " machine chargee a %d / %d, on demande %d en plus : c\'est trop\n",
	//		   placmt->U[m1][r], machine1->C[r], process2->R[r] );
            return FMR_OVERLOAD;
        }
    }

//    DEBUG_PRINT( "Charge de la machine %d avant echange : CPU %d/%d, RAM %d/%d\n",
//		    m2, placmt->U[m2][0], machine2->C[0], placmt->U[m2][1],machine2->C[1] );

    /* Update des charges des ressources de la machine 2 */

    for( r = 0 ; r < instance.nb_ressources ; r++ ) {
        placmt->U[m2][r] +=  process1->R[r] - process2->R[r];
    }

 //   DEBUG_PRINT( "Charge de la machine %d apres echange : CPU %d, RAM %d\n", m2, placmt->U[m2][0], placmt->U[m2][1] );

    /* Update des charges des ressources de la machine 1 */

  //  DEBUG_PRINT( "Charge de la machine %d avant echange : CPU %d, RAM %d\n", m1, placmt->U[m1][0], placmt->U[m1][1] );

    for( r = 0 ; r < instance.nb_ressources ; r++ ) {
        placmt->U[m1][r] += process2->R[r]- process1->R[r] ;
    }

 //   DEBUG_PRINT( "Charge de la machine %d apres echange : CPU %d, RAM %d\n", m1, placmt->U[m1][0], placmt->U[m1][1] );

    /* Ajouter le processus 1 sur la machine 2 */

    listeCompacte_in( & (placmt->liste_procs), m2,(unsigned short)p1);

    /* Retirer le proc 1 de la machine 1 */

    listeCompacte_out( & placmt->liste_procs, m1, (unsigned short)p1);
    /*if (liste_out(p1, & placmt->liste_procs[m1])== LISTE_VIDE) {
	DEBUG_PRINT("erreur: process1 non trouvé\n");
        return FMR_NOT_FOUND;
    }*/

    /* Ajouter le processus 2 sur la machine 1 */

    listeCompacte_in( & (placmt->liste_procs), m1,(unsigned short)p2);

    /* Retirer le proc 2 de la machine 2 */

    listeCompacte_out(& placmt->liste_procs, m2,(unsigned short) p2);
    /*if (liste_out(p2, & placmt->liste_procs[m2]) == LISTE_VIDE) {
	DEBUG_PRINT("erreur: process2 non trouvé\n");
        return FMR_NOT_FOUND;
    }*/

    /* Update de l'assignement */

    placmt->assignment[ p1 ] = (unsigned short) m2;
    placmt->assignment[ p2 ] = (unsigned short)m1;

    return FMR_OK;
}

/* déplace dans le placement fourni en paramètre le proc p vers la machine m_dest
 * les vérifications doivent etre faites avant.
 */
void
deplace_proc(const placement_t const *initial, placement_t *placmt, unsigned short p, unsigned short m_dest) {
	unsigned short r, newmax, i ;
	unsigned short s=instance.procs[p].service;
	unsigned short m_orig=placmt->assignment[p];
	unsigned short l_orig =instance.machines[m_orig].location;
	unsigned short l_dest =instance.machines[m_dest].location;
	unsigned short m0 = initial->assignment[p];
	unsigned short n_orig=instance.machines[m_orig].neighborhood;
	unsigned short n_dest=instance.machines[m_dest].neighborhood;

	if (m_dest ==m_orig)
		fprintf(stderr,"m_dest=m_orig=%d dans deplace_proc!\n",m_dest),exit(1);

	placmt->assignment[p]= (unsigned short)m_dest;
	/* mise à jour du taux de charge, U et residu ne peuvent etre fusionnés */
	for (r=0; r< instance.nb_ressources; r++) {
		if (instance.ressources[r].transcient) {
			if (m0==m_orig) /*  depart de la machine initiale */
				placmt->residu[m0][r] += instance.procs[p].R[r];
			if (m0==m_dest) /* retour sur la machine initiale */
				placmt->residu[m0][r] -= instance.procs[p].R[r];
		}
		placmt->U[m_orig][r] -= instance.procs[p].R[r];
		placmt->U[m_dest][r] += instance.procs[p].R[r];
	}

	listeCompacte_out(& (placmt->liste_procs), m_orig,(unsigned short) p);
	/*if ( liste_out(p, & (placmt->liste_procs[m_orig])) == LISTE_VIDE) {
		DEBUG_PRINT("erreur p non trouvé dans deplace_proc\n");
		exit(FMR_NOT_FOUND);
	}*/
	listeCompacte_in(  &(placmt->liste_procs), m_dest,(unsigned short)p);

     	placmt->nb_procs_par_mach[m_orig]--;
     	placmt->nb_procs_par_mach[m_dest]++;
     	placmt->spread[s][l_orig]--;
	placmt->spread[s][l_dest]++;
	placmt->service_par_machine[m_orig][s]= NOPROCESS;
	placmt->service_par_machine[m_dest][s]= (unsigned short)p;

	if (m_dest == m0) {
		if (placmt->cardServiceMove[s] == 0)
			fprintf(stderr,"erreur cardServiceMove %d de %d à %d (%d)\n",p, m_orig, m_dest, s),exit(1);
		placmt->cardServiceMove[s] --;
		if (placmt->maxCardServiceMove== 1+ placmt->cardServiceMove[s]) {/* zut il faut recalculer le max */
			newmax=0;
			for(i=0; i<instance.nb_services;i++) {
				if (placmt->cardServiceMove[i]> newmax)
					newmax= placmt->cardServiceMove[i];
			}
			placmt->maxCardServiceMove= newmax;
		}
	} else if (m_orig == m0) {
		placmt->cardServiceMove[s] ++;
		if (placmt->maxCardServiceMove < placmt->cardServiceMove[s])
			placmt->maxCardServiceMove = placmt->cardServiceMove[s];
	}
	/* else il avait déjà bougé ! */

	/* mise à jour voisinage_par_service */
	if (n_orig != n_dest) {
		//if ( liste_out(p, & (placmt->liste_de_processus_d_un_service_dans_un_voisinage[n_orig][s])) == LISTE_VIDE)
			//fprintf(stderr,"ps %d pas dans le voisinage n=%d, s=%d!\n",p,n_orig, s),exit(1);
		listeCompacte2D_out(& (placmt->liste_de_processus_d_un_service_dans_un_voisinage), n_orig,s, p);
		listeCompacte2D_in (& (placmt->liste_de_processus_d_un_service_dans_un_voisinage), n_dest,s, p);
		//liste_in(p, & (placmt->liste_de_processus_d_un_service_dans_un_voisinage[n_dest][s]));
	}
}

/* swap_proc version 2 !
 * les vérifications doivent etre faites avant.
 */
int
echange_procs(const placement_t const *initial, placement_t *placmt, unsigned short p1, unsigned short m1, unsigned short p2, unsigned short m2) {
	unsigned short r, newmax, i ;
	unsigned short s1=instance.procs[p1].service;
	unsigned short s2=instance.procs[p2].service;
	unsigned short l1 =instance.machines[m1].location;
	unsigned short l2 =instance.machines[m2].location;
	unsigned short m01 = initial->assignment[p1];
	unsigned short m02 = initial->assignment[p2];
	unsigned short n1=instance.machines[m1].neighborhood;
	unsigned short n2=instance.machines[m2].neighborhood;
	bool recalculMax;

	if (m1 ==m2)
		fprintf(stderr,"m1=m2=%d dans echange_procs!\n",m1),exit(1);

	placmt->assignment[p1]=(unsigned short)m2;
	placmt->assignment[p2]=(unsigned short)m1;

		/* mise à jour du taux de charge, U et residu ne peuvent etre fusionnés */
	for (r=0; r< instance.nb_ressources; r++) {
		if (instance.ressources[r].transcient) {
		       	if (m01==m1)	/* p1 quitte sa machine initiale */
			  	placmt->residu[m01][r] += instance.procs[p1].R[r];
			if (m01==m2){    /* p1 retourne sur sa machine initiale */
#ifdef DEBUG
				if (placmt->residu[m01][r]<instance.procs[p1].R[r]) {
					fprintf(stderr,"residu de %d[%d:%d]trop faible dans echange_procs!\n",
							m01,r,placmt->residu[m01][r]);
					exit(1);
				}
#endif

				placmt->residu[m01][r] -=  instance.procs[p1].R[r];
			}
			if (m02==m2)    /* p2 quitte sa machine initiale */
			  	placmt->residu[m02][r] += instance.procs[p2].R[r];
			if (m02==m1){    /* p2 retourne sur sa machine initiale */
#ifdef DEBUG
				if (placmt->residu[m02][r]<instance.procs[p2].R[r]) {
					fprintf(stderr,"residu de %d[%d:%d]trop faible dans echange_procs!\n",
							m02,r,placmt->residu[m02][r]);
					exit(1);
				}
#endif

				placmt->residu[m02][r] -=  instance.procs[p2].R[r];
			}
		}
		placmt->U[m1][r] += instance.procs[p2].R[r] - instance.procs[p1].R[r];
		placmt->U[m2][r] += instance.procs[p1].R[r] - instance.procs[p2].R[r];
	}

	listeCompacte_out( & (placmt->liste_procs), m1,(unsigned short) p1);
	/*if (liste_out(p1, & (placmt->liste_procs[m1])) == LISTE_VIDE) {
		DEBUG_PRINT("erreur p1=%d sur %d non trouvé dans echange_procs\n", p1, m1);
		exit(FMR_NOT_FOUND);
	}*/
	listeCompacte_in( &(placmt->liste_procs), m2,(unsigned short)p1);
	listeCompacte_out(&(placmt->liste_procs), m2,(unsigned short) p2);

	/*if (liste_out(p2, & (placmt->liste_procs[m2])) == LISTE_VIDE) {
		DEBUG_PRINT("erreur p2=%d sur %d non trouvé dans echange_procs\n", p2, m2);
		exit(FMR_NOT_FOUND);
	}*/
	listeCompacte_in( &(placmt->liste_procs), m1,(unsigned short)p2);

     	placmt->spread[s1][l1]--;
     	placmt->spread[s1][l2]++;
     	placmt->spread[s2][l2]--;
     	placmt->spread[s2][l1]++;
	placmt->service_par_machine[m1][s1]= NOPROCESS;
	placmt->service_par_machine[m2][s2]= NOPROCESS;
	placmt->service_par_machine[m1][s2]= (unsigned short)p2;
	placmt->service_par_machine[m2][s1]= (unsigned short)p1;

	recalculMax=false;
	if (m2 == m01) { /* p1 retourne sur sa machine initiale */
		if (placmt->cardServiceMove[s1] == 0)
			fprintf(stderr,"erreur cardServiceMove %d de %d à %d (service %d)\n",p1, m1, m2, s1),exit(1);
		placmt->cardServiceMove[s1] --;
		if (placmt->maxCardServiceMove== 1+ placmt->cardServiceMove[s1]) {/* zut il faut recalculer le max */
			recalculMax=true;
		}
	} else if (m1 == m01) { /* p1 quitte de sa machine initiale */
		placmt->cardServiceMove[s1] ++;
		if (placmt->maxCardServiceMove < placmt->cardServiceMove[s1]) {
			placmt->maxCardServiceMove = placmt->cardServiceMove[s1];
		}
	}
	if (m1 == m02) { /* p2 retourne sur sa machine initiale */
		if (placmt->cardServiceMove[s2] == 0)
			fprintf(stderr,"erreur cardServiceMove %d de %d à %d (service %d)\n",p2, m2, m1, s2),exit(1);
		placmt->cardServiceMove[s2] --;
		if (placmt->maxCardServiceMove== 1+ placmt->cardServiceMove[s2]) {/* zut il faut recalculer le max */
			recalculMax=true;
		}
	} else if (m2 == m02) { /* p2 quitte sa machine initiale */
		placmt->cardServiceMove[s2] ++;
		if (placmt->maxCardServiceMove < placmt->cardServiceMove[s2]) {
			placmt->maxCardServiceMove = placmt->cardServiceMove[s2];
		}
	}
	recalculMax = recalculMax && placmt->maxCardServiceMove != placmt->cardServiceMove[s1] && placmt->maxCardServiceMove != placmt->cardServiceMove[s2] ;
	if ( recalculMax ) {
		newmax=0;
		for(i=0; i<instance.nb_services;i++) {
			if (placmt->cardServiceMove[i]> newmax)
				newmax= placmt->cardServiceMove[i];
		}
		placmt->maxCardServiceMove= newmax;
	}

	/* mise à jour voisinage_par_service */
	if (n1 != n2) {
		//if ( liste_out(p1, & placmt->liste_de_processus_d_un_service_dans_un_voisinage[n1][s1]) == LISTE_VIDE)
			//fprintf(stderr,"ps %d pas dans le voisinage n=%d, s=%d!\n",p1,n1, s1),exit(1);
		listeCompacte2D_out(& placmt->liste_de_processus_d_un_service_dans_un_voisinage, n1,s1, p1);
		listeCompacte2D_in (& placmt->liste_de_processus_d_un_service_dans_un_voisinage, n2,s1, p1);
		//liste_in(p1, & placmt->liste_de_processus_d_un_service_dans_un_voisinage[n2][s1]);
		//if ( liste_out(p2, & placmt->liste_de_processus_d_un_service_dans_un_voisinage[n2][s2]) == LISTE_VIDE)
			//fprintf(stderr,"ps %d pas dans le voisinage n=%d, s=%d!\n",p2,n2, s2),exit(1);
		//liste_in(p2, & placmt->liste_de_processus_d_un_service_dans_un_voisinage[n1][s2]);
		listeCompacte2D_out(& placmt->liste_de_processus_d_un_service_dans_un_voisinage, n2,s2, p2);
		listeCompacte2D_in (& placmt->liste_de_processus_d_un_service_dans_un_voisinage, n1,s2, p2);
	}
	return FMR_OK;
}

/* enlever le processus p du placement, suppose que les vérifs de contraintes
   on déjà été faites, voir check_remove_proc */
void
oter_proc_glouton (const placement_t const *initial, unsigned short p, unsigned short m,
	   placement_t * placmt)
{
    unsigned short i, r;
    unsigned short s = instance.procs[p].service;
    unsigned short l = instance.machines[m].location;
    unsigned short m0 = initial->assignment[p];
    unsigned short n = instance.machines[m].neighborhood;
    unsigned short newmax;

     placmt->assignment[p]=NOMACHINE;
    /* mise à jour du taux de charge, U et residu ne peuvent etre fusionnés */
    for (r = 0; r < instance.nb_ressources; r++) {
/*	if (m== m0 && instance.ressources[r].transcient )
		placmt->residu[m][r] += instance.procs[p].R[r];*/
	placmt->U[m][r] -= instance.procs[p].R[r];
    }

    listeCompacte_out(&(placmt->liste_procs), m,(unsigned short) p);
    /*if (liste_out (p, &(placmt->liste_procs[m])) == LISTE_VIDE) {
	DEBUG_PRINT ("erreur p non trouvé dans ote_proc\n");
	exit (FMR_NOT_FOUND);
    }*/

    placmt->nb_procs_par_mach[m]--;
    placmt->spread[s][l]--;
    placmt->service_par_machine[m][s] = NOPROCESS;

    if (m != m0) {
	placmt->cardServiceMove[s]--;
	if (placmt->maxCardServiceMove == placmt->cardServiceMove[s] + 1) {	/* donc il faut recalculmax */
	    //recalcul_maxCardServiceMove(initial, placmt, s, -1);
	    newmax = 0;
	    for (i = 0; i < instance.nb_services; i++) {
		if (placmt->cardServiceMove[i] > newmax)
		    newmax = placmt->cardServiceMove[i];
	    }
	    placmt->maxCardServiceMove = newmax;
	}
    }

    /* mise à jour voisinage_par_service */
    listeCompacte2D_out(&(placmt->liste_de_processus_d_un_service_dans_un_voisinage),n,s,p);
}


void
ajouter_proc_glouton (const placement_t const *initial, unsigned short p, unsigned short m, placement_t * placmt)
{
    unsigned short r;
    unsigned short s = instance.procs[p].service;
    unsigned short l = instance.machines[m].location;
    unsigned short m0 = initial->assignment[p];
    unsigned short n = instance.machines[m].neighborhood;

     placmt->assignment[p]= (unsigned short)m;
    /* mise à jour du taux de charge, U et residu ne peuvent etre fusionnés */
    for (r = 0; r < instance.nb_ressources; r++) {
        if (m0 !=m && instance.ressources[r].transcient) {
                placmt->residu[m0][r] += instance.procs[p].R[r];
        }
        placmt->U[m][r] += instance.procs[p].R[r];
    }

    listeCompacte_in( &(placmt->liste_procs), m,(unsigned short)p);

    placmt->nb_procs_par_mach[m]++;
    placmt->spread[s][l]++;
    placmt->service_par_machine[m][s] = (unsigned short)p;

    if (m != m0) {
        if (placmt->maxCardServiceMove == placmt->cardServiceMove[s]) {
            placmt->maxCardServiceMove++;
        }
        placmt->cardServiceMove[s]++;
    }
    //liste_in (p, &(placmt->liste_de_processus_d_un_service_dans_un_voisinage[n][s]));
	listeCompacte2D_in(  &(placmt->liste_de_processus_d_un_service_dans_un_voisinage), n, s, p);
}
