/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "fmr.h"
#include "fmr_mem.h"
#include "fmr_cost.h"
#include "fmr_tools.h"
#include "fmr_exploRandSwap.h"
#include "fmr_exploShift.h"
#include "fmr_exploSwap.h"
#include "fmr_exploEjection.h"
#include "fmr_recuit.h"
#include "fmr_recuit_b.h"
#include "fmr_glouton.h"
#include "fmr_core.h"
#include "fmr_threads.h"
#include "fmr_relax.h"


/**
 * Le calcul lui-meme par la relax vns.
 * Prend en entree un pointeur vers l'instance, un pointeur vers
 * le placement initial et un pointeur vers le resultat du calcul.
 * Retourne le gain obtenu.
 */

int64
calcul_assignement_relax (const placement_t * initial_relax, placement_t* resultat, unsigned *seed) {
    int64 gain=0;
    unsigned int i,j;
    long gainMin = 0;		/* a parametrer dynamiquement */
    unsigned short longChaine = MAXLONGCHAINEEJECTION;	/* choisi par experimentation */
    placement_t resultat_courant;
    struct drand48_data rand_buf ;
    // algos. '.'=retour à "initial"  '_'=retour à "resultat"
//     char algos[] = "._EXWSBYCT";
    char algos[] = "EXWSBYC";
//     enum {E=2, X, W, S, B, Y, C, T};
    enum {E=0, X, W, S, B, Y, C};
    size_t NbAlgos = strlen(algos);
    double weight[NbAlgos];	/*Poids des algos dans l'ordre: reinit, C, E, S, B, W que l'on peut utiliser dans la VNS */
    double max_fitness;
    int pi[NbAlgos];	/*Score des algos que l'on a utilise dans la VNS */
    int theta[NbAlgos];	/*Nombre de tentatives dans la VNS pour chaque algo */
    double random;
    double running_total[NbAlgos];
    double current_total;
    double r=0.2;
    int reinit=0;
    unsigned int arret = 0;
    resultat_courant = alloc_assignement(instance.nb_procs);
    
    copier_assignement(resultat, &resultat_courant);

    /*Initialisation des poids et scores des algos*/
    for(i=0;i<NbAlgos;i++) {
      weight[i]=pi[i]=theta[i]=1;
    }
    
    for(j=0;j<10;j++) {
      /* Ajustement dynamique des poids en fonction de leur score */
      for (i=0; i<NbAlgos; i++) {
        weight[i] = weight[i]*(1-r) + r*((double)pi[i]/theta[i]);
//         DEBUG_PRINT ("VNS: weight[%d]=%lf\n", i, weight[i]);
      }
    
      /* Roulette wheel selection*/

      current_total = 0;
 
      for (i=0; i<NbAlgos; i++) {
        current_total +=   weight[i];
        running_total[i] = current_total;
//         DEBUG_PRINT ("VNS: running_total[%d]=%lf\n", i, running_total[i]);
      }
      max_fitness = current_total;
 
      random = max_fitness*rand_r(seed)/RAND_MAX;
      if(arret==50) random=0.001;
//      if(arret==50) 
//      {
//                resultat=atomic_get_meilleur_placement();
//		copier_assignement(resultat,  &resultat_courant );
//      }
//       DEBUG_PRINT ("VNS: random=%lf\n", random);
      for (i=0; i<NbAlgos; i++) {
	if (random > running_total[i]) {
	  //on ne fait rien
	}
	else
	{
          switch (algos[i]) {
/* 	    case '.': // retour à la solution "initiale"
		copier_assignement(initial_relax,  &resultat_courant );
		theta[i]++;
		arret++;
		gain=0;
		DEBUG_PRINT ("VNS: theta[0]=%d et pi[0]=%d\n", theta[0], pi[0]);
		break;
 	    case '_': // retour à la solution "resultat"
		copier_assignement(resultat,  &resultat_courant );
		theta[1]++;
		arret++;
		gain=0;
		DEBUG_PRINT ("VNS: theta[1]=%d et pi[1]=%d\n", theta[1], pi[1]);
		break;
 	    case 'T': // retour à la meilleure solution
                atomic_copy_meilleur_placement(&resultat_courant);
		theta[T]++;
		arret++;
		gain=0;
		DEBUG_PRINT ("VNS: theta[T]=%d et pi[T]=%d\n", theta[T], pi[T]);
		break;*/
            case 'E': //E
                DEBUG_PRINT ("Explo%d: SwapVNS relax (%ld)\n", i, gainMin);
                gain = explo_voisinageSwap (initial_relax, &resultat_courant, gainMin);	//Echange
		theta[E]++;
		if (gain > gainMin)
		{
		  pi[E]++;
		  arret=0;
		}
		else
		  arret++;
		DEBUG_PRINT ("VNS: theta[E]=%d et pi[E]=%d\n", theta[E], pi[E]);
                break;
            case 'X': //X
                DEBUG_PRINT ("Explo%d: VeryBestswapVNS relax (%ld)\n", i, gainMin);
                gain = explo_voisinageVeryBestSwap(initial_relax, &resultat_courant, gainMin);	//Echange
		theta[X]++;
		if (gain > gainMin)
		{
		  pi[X]++;
		  arret=0;
		}
		else
		  arret++;
		DEBUG_PRINT ("VNS: theta[X]=%d et pi[X]=%d\n", theta[X], pi[X]);
                break;
            case 'W': //W
                DEBUG_PRINT ("Explo%d: BestswapVNS relax (%ld)\n", i, gainMin);
                gain = explo_voisinageBestSwap(initial_relax, &resultat_courant, gainMin);	//Echange
		theta[W]++;
		if (gain > gainMin)
		{
		  pi[W]++;
		  arret=0;
		}
		else
		  arret++;
		DEBUG_PRINT ("VNS: theta[W]=%d et pi[W]=%d\n", theta[W], pi[W]);
                break;
            case 'S': // S
                DEBUG_PRINT ("Explo%d: ShiftVNS relax (%ld)\n", i, gainMin);	// Shift
                gain = explo_voisinageShift (initial_relax, &resultat_courant, gainMin);
		theta[S]++;
		if (gain > gainMin)
		{
		  pi[S]++;
		  arret=0;
		}
		else
		  arret++;
		DEBUG_PRINT ("VNS: theta[S]=%d et pi[S]=%d\n", theta[S], pi[S]);
                break;
            case 'B': // B
                DEBUG_PRINT ("Explo%d: BestShiftVNS relax (%ld)\n", i, gainMin);
                gain = explo_voisinageBestShift (initial_relax, &resultat_courant, gainMin);	//Best shift
		theta[B]++;
		if (gain > gainMin)
		{
		  pi[B]++;
		  arret=0;
		}
		else
		  arret++;
		DEBUG_PRINT ("VNS: theta[B]=%d et pi[B]=%d\n", theta[B], pi[B]);
                break;
            case 'Y': // Y
                DEBUG_PRINT ("Explo%d: VeryBestShiftVNS relax (%ld)\n", i, gainMin);
                gain = explo_voisinageVeryBestShift (initial_relax, &resultat_courant, gainMin);	//Best shift
		theta[Y]++;
		if (gain > gainMin)
		{
		  pi[Y]++;
		  arret=0;
		}
		else
		  arret++;
		DEBUG_PRINT ("VNS: theta[Y]=%d et pi[Y]=%d\n", theta[Y], pi[Y]);
                break;
            case 'C': // C
                DEBUG_PRINT ("Explo%d: EjectionChainVNS relax (%ld)\n", i, gainMin);
                gain = explo_voisinageEjectionChain (initial_relax, &resultat_courant, longChaine,
                                                   gainMin);
		theta[C]++;
		if (gain > gainMin)
		{  
		  pi[C]++;
		  arret=0;
		}
		else
		  arret++;
		DEBUG_PRINT ("VNS: theta[C]=%d et pi[C]=%d\n", theta[C], pi[C]);
                break;
           default:
                fprintf (stderr, "Erreur algo inconnu\n");
		gain=0;
                break;
          }
          if (gain>gainMin)
	      /* on suppose que le résultat courant n'a pas été modifié
	       * dans le cas contraire */
	      atomic_set_meilleur_placement(&resultat_courant, algos[i]);
	  break;
	}
      }
      if (alarme)
        return FMR_TIMEOUT;
      if(arret>200) break;
    }
    
    if (alarme)
      return FMR_TIMEOUT;
    
    gain = (long) (resultat->cout - resultat_courant.cout);
    free_placement( &resultat_courant );
    return gain;
}

