/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/

#include "fmr.h"
#include "fmr_servicegraph.h"


void construireGraphDependances( graphnode_t* graphTab ){

    unsigned int s, d;  /* s = service, d = dependance */
    unsigned int* compteur;

    /* On garde tous les noeuds lineairement dans un tableau
       et on a des liens unidirectionnels entre les noeuds
    */
    
    compteur = (unsigned int*) malloc( instance.nb_services * sizeof( unsigned int ) );

    /* Premier tour : on compte les services qui dependent de chaque service 
       (liens down)
    */
    for( s = 0 ; s < instance.nb_services ; s++ ) {

        /* Initialisation */

        graphTab[s].nb_dep_down = 0;
        compteur[s] = 0;
    }
        /* On parcourt tout le monde.... */
    for( s = 0 ; s < instance.nb_services ; s++ ) {
        for( d = 0 ; d < instance.services[s].nb_dep ; d++ ) {
	  unsigned int dep;
	  dep = instance.services[s].dependances[d];
	  graphTab[dep].nb_dep_down++;
        }
    }

   /* Deuxieme tour : allocations memoire... */

    for( s = 0 ; s < instance.nb_services ; s++ ) {
        /*if( LIKELY( 0 != graphTab[s].nb_dep_down ) ) {*/
        if(  0 != graphTab[s].nb_dep_down  ) {
            graphTab[s].services_down = (graphnode_t**) malloc( graphTab[s].nb_dep_down * sizeof( graphnode_t *) );
        } else  {
            graphTab[s].services_down = NULL;
        }
    }

   /* Troisieme tour : remplissage */

    for( s = 0 ; s < instance.nb_services ; s++ ) {
        unsigned int dep;
        for( d = 0 ; d < instance.services[s].nb_dep ; d++ ) {
            dep = instance.services[s].dependances[d];
            graphTab[s].services_down[compteur[s]] = &( graphTab[dep] );
            compteur[s]++;
        }
    }

   /* Quatrieme tour : on remplit les services dont on depend 
       (liens up) et l'id (qui n'est qu'un raccourcis car 
       c'est en fait l'indice du noeud dans le tableau de noeuds)
    */

    for( s = 0 ; s < instance.nb_services ; s++ ) {
        graphTab[s].id = s;  
        graphTab[s].nb_dep_up = instance.services[s].nb_dep;
        if( LIKELY( 0 != instance.services[s].nb_dep ) ) {
            graphTab[s].services_up = (graphnode_t**) malloc( graphTab[s].nb_dep_up * sizeof( graphnode_t*) );
            for( d = 0 ; d < graphTab[s].nb_dep_up ; d++ ) {
                graphTab[s].services_up[d] = &( graphTab[instance.services[s].dependances[d]] );
            }
        } else  {
            graphTab[s].services_up = NULL;
        }
        
    }

}

/* Determine si un service est un puis
   c'est-a-dire si il n'existe pas de services qui dependent de lui
 */

bool estPuis( graphnode_t* graphTab, int num ) {
    if( NULL == graphTab[num].services_down ) {
        return true;
    } else {
        return false;
    }
}

/* Determine si un service est une source
   c'est-a-dire si il depend d'au moins un service
 */

bool estSource( graphnode_t* graphTab, int num ) {
    if( NULL == graphTab[num].services_up ) {
        return true;
    } else {
        return false;
    }
}

