/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/

#ifndef _FMR_TIRAGE_H
#define _FMR_TIRAGE_H

typedef enum {
    SHIFT = 0,
    SWAP
} tirage_type_t;



#include "fmr.h"

/**
* tirages aléatoires des mouvements
*/

typedef struct tirage_ {
    tirage_type_t type ;
    bool nouveau;
    union {
        struct {
            unsigned short p;
            unsigned short m;
        } shift;
        struct {
            unsigned short p1;
            unsigned short p2;
        } swap;
    } t;
    int64 gain;
    #ifdef EXAMINE_RECUIT
    int64 nb_essais ;
    #endif
} tirage_t;


/// nouveau tirage. On fait jusqu'à nb_essais jusqu'à trouver un mouvement possible
tirage_t tirage_shift_checked(placement_t* placement, unsigned long nb_essais, struct drand48_data * rand_buf);
tirage_t tirage_swap_checked(placement_t* placement, unsigned long nb_essais, struct drand48_data * rand_buf);
tirage_t tirage_next(placement_t* placement, unsigned long nb_essais, struct drand48_data * rand_buf);

/// nouveau tirage. On fait jusqu'à t>tmax jusqu'à trouver un mouvement possible
tirage_t tirage_next_b(placement_t* placement, double tmax, struct drand48_data * rand_buf);

/// nouveau tirage. On fait jusqu'à t>tmax_essais jusqu'à trouver un mouvement possible
tirage_t tirage_next_c(const placement_t* initial, placement_t* placement, double tmax, struct drand48_data * rand_buf, tirage_type_t type);

/// calcule le gain d'un mouvement obtenu par tirage aléatoire
int64 calcule_gain_tirage( const placement_t* initial, placement_t* courant, tirage_t tirage );


/// effectue le(s) déplacement(s) d'un tirage aléatoire
void effectue_tirage( const placement_t * initial, placement_t * courant, tirage_t tirage );


#ifdef DEBUG
void DBG_AFFICHE_TIRAGE(tirage_t tirage, int64 cout);
#else
#define DBG_AFFICHE_TIRAGE(...)
#endif


#ifdef EXAMINE_RECUIT
# define PRINT_EXAMINE_RECUIT(...) \
{ static double _next_t=0; double _t=tempsEcoule2(); \
  if (_t>=_next_t) {_next_t=_t+1; printf(__VA_ARGS__); } }
#else
# define PRINT_EXAMINE_RECUIT(...) 
#endif


#endif
